# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as python3
import copy
import enum
from typing import Any

import numpy as np
import pyspiel

_NUM_PLAYERS = 2
_NUM_MAX_PEACE_STEP = 10
_NUM_ROWS = 6
_NUM_COLS = 3
_NUM_CELLS = _NUM_ROWS * _NUM_COLS
_NUM_CHESS_TYPES = 6
_DICT_CHESS_CELL = {9: 6, 8: 5, 7: 4, 6: 3, 2: 2, 1: 1, 0: 0}
_DICT_CHESS_NAME = {6: "雷", 5: "师", 4: "旅", 3: "团", 2: "炸", 1: "旗", 0: "空", -10:"口"}

_GAME_TYPE = pyspiel.GameType(
    short_name="junqi1_py38",
    long_name="Python Simplized Junqi1",
    dynamics=pyspiel.GameType.Dynamics.SEQUENTIAL,
    chance_mode=pyspiel.GameType.ChanceMode.DETERMINISTIC,
    information=pyspiel.GameType.Information.IMPERFECT_INFORMATION,
    utility=pyspiel.GameType.Utility.ZERO_SUM,
    reward_model=pyspiel.GameType.RewardModel.TERMINAL,
    max_num_players=_NUM_PLAYERS,
    min_num_players=_NUM_PLAYERS,
    provides_information_state_string=True,
    provides_information_state_tensor=False,
    provides_observation_string=True,
    provides_observation_tensor=True,
    provides_factored_observation_string=True)  #

_GAME_INFO = pyspiel.GameInfo(
    num_distinct_actions=_NUM_CELLS * _NUM_CELLS + 1,
    max_chance_outcomes=0,  #
    num_players=2,
    min_utility=-1.0,
    max_utility=1.0,
    utility_sum=0.0,
    max_game_length=200  # _NUM_CELLS * _NUM_CELLS
)


class GamePhase(enum.IntEnum):
    """Enum game phrase."""
    DEPLOYING: int = 0
    SELECTING: int = 1
    MOVING: int = 2


class ChessType(enum.IntEnum):
    """Enum chess type to make the code easy-reading."""
    MINE: int = 6
    GENERAL: int = 5
    COLONEL: int = 4
    CAPTAIN: int = 3
    BOMB: int = 2
    FLAG: int = 1
    NONE: int = 0
    UNKOWN: int = -10


class JunQiGame(pyspiel.Game):
    """A Python version of the JunQi game."""

    def __init__(self, params=None):
        super().__init__(_GAME_TYPE, _GAME_INFO, params or dict())

    def new_initial_state(self):
        """Returns a state corresponding to the start of a game."""
        return JunQiState(self)

    def make_py_observer(self, iig_obs_type=None, params=None):
        """Returns an object used for observing game state."""
        return JunQiObserver(
            iig_obs_type or pyspiel.IIGObservationType(perfect_recall=False),
            params)


class JunQiState(pyspiel.State):
    """A python version of the JunQi state."""

    def __init__(self, game):
        """Constructor; should only be called by Game.new_initial_state."""
        super().__init__(game)

        self._cur_player: int = 0
        self._player0_score: float = 0.0
        self._is_terminal: bool = False

        self.game_phase: GamePhase = GamePhase.DEPLOYING
        self.game_length_real: int = 0
        self.game_length_peace: int = 0
        self.game_length: int = 0

        self.selected_pos = np.array([[0, 0], [_NUM_ROWS - 1, _NUM_COLS - 1]])
        self.decode_action = np.array([0, 0] * (_NUM_COLS * _NUM_ROWS))
        self.board = np.array([[Chess(0, -1)] * _NUM_COLS for _ in range(_NUM_ROWS)])
        self.chess_list = [[6, 5, 5, 4, 4, 3, 3, 2, 1],
                                            [6, 5, 5, 4, 4, 3, 3, 2, 1]]
        self.obs_mov = [[[0] * _NUM_COLS for _ in range(_NUM_ROWS)],
                                               [[0] * _NUM_COLS for _ in range(_NUM_ROWS)]]
        self.obs_attack: bool = False

        for i in range(_NUM_COLS * _NUM_ROWS):
            self.decode_action[i] = [i // _NUM_COLS, i % _NUM_COLS]
        pass

    # OpenSpiel (PySpiel) API functions are below. This is the standard set that
    # should be implemented by every perfect-information sequential-move game.

    def current_player(self) -> int or any:
        """Returns id of the next player to move, or TERMINAL if game is over."""
        return pyspiel.PlayerId.TERMINAL if self._is_terminal else self._cur_player

    def _legal_actions(self, player: int):
        """Returns a list of legal actions."""
        actions: list[bool] = [False] * (_NUM_COLS * _NUM_ROWS)
        actions_idx_list: list[int] = []
        if self.game_phase == GamePhase.DEPLOYING:
            for i in range(_NUM_COLS * _NUM_ROWS // 2):
                actions[i] = True if self.chess_list[player][i] != 0 else False
        elif self.game_phase == GamePhase.SELECTING:
            for i in range(_NUM_ROWS):
                for j in range(_NUM_COLS):
                    if (self.board[i][j].country == player
                            and self.board[i][j].type != ChessType.MINE
                            and self.board[i][j].type != ChessType.FLAG
                            and self._is_legal_start([i, j], player)):
                        actions[i * _NUM_COLS + j] = True
        elif self.game_phase == GamePhase.MOVING:
            from_pos: list[int, int] = self.selected_pos[player]
            for to_pos in [[from_pos[0] + 1, from_pos[1]],
                           [from_pos[0] - 1, from_pos[1]],
                           [from_pos[0], from_pos[1] + 1],
                           [from_pos[0], from_pos[1] - 1]]:
                if self._is_legal_destination(to_pos, player):
                    actions[to_pos[0] * _NUM_COLS + to_pos[1]] = True
        for i in range(len(actions)):
            if actions[i]:
                actions_idx_list.append(i)
        if len(actions_idx_list) == 0:
            actions_idx_list.append(_NUM_COLS * _NUM_ROWS)
        return actions_idx_list

    def _is_legal_destination(self, to_pos: list[int, int], player: int) -> bool:
        """Check whether the destination of a move is legal."""
        r, c = to_pos[0], to_pos[1]
        return True if (0 <= r < _NUM_ROWS and 0 <= c < _NUM_COLS
                        and self.board[r][c].country != player) else False

    def _is_legal_start(self, from_pos: list[int, int], player: int) -> bool:
        """Check whether the start point of a move is legal."""
        for to_pos in [[from_pos[0] + 1, from_pos[1]],
                       [from_pos[0] - 1, from_pos[1]],
                       [from_pos[0], from_pos[1] + 1],
                       [from_pos[0], from_pos[1] - 1]]:
            if self._is_legal_destination(to_pos, player):
                return True
        return False

    def _apply_action(self, action: int) -> None:
        """Applies the specified action to the state."""
        # TODO: Remove copy module if we can, and refactor the code to easier ones.
        # TODO: Try to make the code easier
        # print(self.serialize(), end="\n\n") if self.game_phase != GamePhase.SELECTING else print("", end="")
        player = self._cur_player

        if action == _NUM_COLS * _NUM_ROWS:
            # End game by no legal move.
            self._is_terminal = True
            self._player0_score = -1.0 if player == 0 else 1.0
            return

        if self.game_phase == GamePhase.DEPLOYING:
            r, c = self.selected_pos[player][0], self.selected_pos[player][1]
            self.board[r][c] = Chess(self.chess_list[player][action], player)
            self.chess_list[player][action] = 0
            if player == 0:
                r += (c + 1) // _NUM_COLS
                c = (c + 1) % _NUM_COLS
            else:
                r += (c - 1) // _NUM_COLS
                c = (c - 1) % _NUM_COLS
                if r == (_NUM_ROWS // 2) - 1 and c == _NUM_COLS - 1:
                    # Deployment phase ended, start selecting-moving phase
                    self.game_phase = GamePhase.SELECTING
            # TODO: Maybe this line is not necessary.
            self.selected_pos[player][0], self.selected_pos[player][1] = r, c
            self._cur_player = 1 - self._cur_player

        elif self.game_phase == GamePhase.SELECTING:
            self.selected_pos[player] = copy.deepcopy(self.decode_action[action])
            self.game_phase = GamePhase.MOVING

        elif self.game_phase == GamePhase.MOVING:
            self.obs_attack = False
            self.obs_mov: list[list[list[int]]] = [[[0] * _NUM_COLS for _ in range(_NUM_ROWS)],
                                                   [[0] * _NUM_COLS for _ in range(_NUM_ROWS)]]

            attacker: Chess = copy.deepcopy(self.board[self.selected_pos[player][0]][self.selected_pos[player][1]])
            defender: Chess = copy.deepcopy(self.board[self.decode_action[action][0]][self.decode_action[action][1]])

            if defender.type == ChessType.NONE:
                self.board[self.decode_action[action][0]][self.decode_action[action][1]] = copy.deepcopy(attacker)

                self.obs_mov[player][self.selected_pos[player][0]][self.selected_pos[player][1]] = -1
                self.obs_mov[1 - player][self.selected_pos[player][0]][self.selected_pos[player][1]] = -1

            else:
                self.obs_attack = True
                self.obs_mov[player][self.selected_pos[player][0]][self.selected_pos[player][1]] = -2
                self.obs_mov[1 - player][self.selected_pos[player][0]][self.selected_pos[player][1]] = -2
                if defender.type == ChessType.FLAG:
                    # End game by captured flag.
                    self._is_terminal = True
                    self._player0_score = 1.0 if self._cur_player == 0 else -1.0
                elif (attacker.type == ChessType.BOMB
                      or defender.type == ChessType.BOMB
                      or attacker.type == defender.type):
                    self.board[self.decode_action[action][0]][self.decode_action[action][1]] = Chess(0, -1)
                elif attacker.type > defender.type:
                    self.board[self.decode_action[action][0]][self.decode_action[action][1]] = copy.deepcopy(attacker)
                elif attacker.type < defender.type:
                    pass

            self.game_length_peace += 1
            if self.obs_attack:
                self.game_length_peace = 0

            self.obs_mov[player][self.decode_action[action][0]][self.decode_action[action][1]] = 1
            self.obs_mov[1 - player][self.decode_action[action][0]][self.decode_action[action][1]] = 1

            self.board[self.selected_pos[player][0]][self.selected_pos[player][1]] = Chess(0, -1)

            self._cur_player = 1 - self._cur_player
            self.game_length_real += 1
            self.game_phase = GamePhase.SELECTING

        self.game_length += 1

    def _action_to_string(self, player, action):
        """Action -> string."""
        to_pos = self.decode_action[action]
        return "({},{})".format("0" if player == 0 else "1", to_pos)

    def is_terminal(self):
        """Returns True if the game is over."""
        return self._is_terminal

    def returns(self):
        """Total reward for each player over the course of the game so far."""
        return [self._player0_score, -self._player0_score]

    def serialize(self):
        return _board_to_string(self.board)

    def serialize_pov(self, player) -> str:
        pov_board: list[list[Chess]] = [[Chess(0, -1)] * _NUM_COLS for _ in range(_NUM_ROWS)]
        for row in range(_NUM_ROWS):
            for col in range(_NUM_COLS):
                if self.board[row][col].country == player:
                    pov_board[row][col] = self.board[row][col]
                elif self.board[row][col].country == 1 - player:
                    pov_board[row][col] = Chess(-10, 1 - player)
        return _board_to_string_pov(pov_board, self.board)

    def serialize_action(self, action):
        return f"Action:{self.decode_action[action]} AKA {action}"

    def __str__(self):
        """String for debug purposes. No particular semantics are required."""
        return _board_to_string(self.board)


class JunQiObserver:
    """Observer, conforming to the PyObserver interface (see observation.py)."""

    def __init__(self, iig_obs_type, params):
        """Initializes an empty observation tensor."""
        if params:
            raise ValueError(f"Observation parameters not supported; passed {params}")
        # The observation should contain a 1-D tensor in `self.tensor` and a
        # dictionary of views onto the tensor, which may be of any shape.
        # Here the observation is indexed `(cell state, row, column)`.
        # TODO: Add interface for the number of moves of history

        self.num_history_move: int = 20
        scalar_shape: int = 1
        prev_select_shape: int = 1

        # Observation components. See paper page 37.
        # Note that we deleted "lakes on the map".
        #
        #                          Private Info.      Public Info. -i    Public Info. i     Move m i
        self.shape = (  # pointers      |                     |                 |               |
            _NUM_ROWS, _NUM_COLS, (_NUM_CHESS_TYPES + _NUM_CHESS_TYPES + _NUM_CHESS_TYPES + self.num_history_move +
                                   scalar_shape + scalar_shape + scalar_shape + scalar_shape + prev_select_shape))
        #                          Remain Len.    Remain Mov.    Game Phase     Pha.Sele.Mov.  Prev. Selectoin

        self.tensor = np.zeros(np.prod(self.shape), np.float32)
        self.dict = {"observation": np.reshape(self.tensor, self.shape)}

        self.mov_idx: int = 0
        idx = 3 * _NUM_CHESS_TYPES + self.num_history_move + 1
        for row in range(_NUM_ROWS):
            for col in range(_NUM_COLS):
                self.dict["observation"][row][col][idx] = _NUM_MAX_PEACE_STEP

    def set_from(self, state, player):
        """Updates `tensor` and `dict` to reflect `state` from PoV of `player`."""
        # We update the observation via the shaped tensor since indexing is more
        # convenient than with the 1-D tensor. Both are views onto the same memory.
        obs = self.dict["observation"]
        prev_obs = copy.deepcopy(obs)
        obs.fill(0)

        _idx = 0
        self.mov_idx = (self.mov_idx + 1) if (self.mov_idx <= self.num_history_move
                                              and state.game_phase == GamePhase.SELECTING) else 0

        for row in range(_NUM_ROWS):
            for col in range(_NUM_COLS):
                chess = state.board[row][col]

                # The player’s own private information.
                # Shape: _NUM_ROWS * _NUM_COLS * _NUM_CHESS_TYPES tensor.
                _idx = 0
                if chess.country == player:
                    for t in range(1, _NUM_CHESS_TYPES + 1):
                        if chess.type == t:
                            obs[row][col][t - 1] = 1

                # The opponent’s public information.
                # Contains all 0’s during the deployment phase.
                # Shape: _NUM_ROWS * _NUM_COLS * _NUM_CHESS_TYPES tensor.
                # TODO: #unrevealed(t) in paper page 36.
                _idx = _NUM_CHESS_TYPES
                if chess.country == 1 - player:
                    for i in range(_NUM_CHESS_TYPES):
                        obs[row][col][i + _idx] = 2
                    for t in range(self.num_history_move):
                        if prev_obs[row][col][t + 3 * _NUM_CHESS_TYPES] == 1:
                            for i in range(_NUM_CHESS_TYPES):
                                obs[row][col][i + _idx] = 3
                                break
                # ...
                # TODO: Add the situation "if the piece at (r, c) is known to have type t".
                #       For example 40 died and then the position of the flag
                #       should be a public information.
                # ...
                # To be continued.

                # The player’s own public information.
                # Contains all 0’s during the deployment phase.
                # Shape: _NUM_ROWS * _NUM_COLS * _NUM_CHESS_TYPES tensor.
                _idx = 2 * _NUM_CHESS_TYPES
                if chess.country == player:
                    for i in range(_NUM_CHESS_TYPES):
                        obs[row][col][i + _idx] = 2
                    for t in range(self.num_history_move):
                        if prev_obs[row][col][t + 3 * _NUM_CHESS_TYPES] == 1:
                            for i in range(_NUM_CHESS_TYPES):
                                obs[row][col][i + _idx] = 3
                                break
                # ...
                # TODO: Add the situation "if the piece at (r, c) is known to have type t".
                #       For example 40 died and then the position of the flag
                #       should be a public information.
                # ...
                # To be continued.

                # An encoding of the last 40(or other number) moves.
                # Here we used a scrolling index.
                # Shape: _NUM_ROWS * _NUM_COLS * self.num_history_move tensor.
                # TODO: Need changes in state
                _idx = 3 * _NUM_CHESS_TYPES
                obs[row][col][self.mov_idx + _idx] = copy.deepcopy(state.obs_mov[player][row][col])
                # print(obs[row][col][self.mov_idx + _idx])

                # The ratio of the game length to the maximum length
                # before the game is considered a draw.
                # Shape: Scalar -> _NUM_COLS * _NUM_ROWS * 1 tensor.
                _idx = 3 * _NUM_CHESS_TYPES + self.num_history_move
                obs[row][col][_idx] = _GAME_INFO.max_game_length - state.game_length

                # The ratio of the number of moves since the last attack
                # to the maximum number of moves without attack before
                # the game is considered a draw.
                # Shape: Scalar -> _NUM_COLS * _NUM_ROWS * 1 tensor.
                _idx = 3 * _NUM_CHESS_TYPES + self.num_history_move + 1
                obs[row][col][_idx] = _NUM_MAX_PEACE_STEP - state.game_length_peace

                # The phase of the game.
                # Either deployment (1) or play (0).
                # Shape: Scalar -> _NUM_COLS * _NUM_ROWS * 1 tensor.
                _idx = 3 * _NUM_CHESS_TYPES + self.num_history_move + 2
                obs[row][col][_idx] = 1 if state.game_phase == GamePhase.DEPLOYING else 0
                _idx += 1

                # An indication of whether the agent needs to select a
                # piece (0) or target square (1) for an already selected piece.
                # 0 during deployment phase.
                # Shape: Scalar -> _NUM_COLS * _NUM_ROWS * 1 tensor.
                _idx = 3 * _NUM_CHESS_TYPES + self.num_history_move + 3
                obs[row][col][_idx] = 1 if state.game_phase == GamePhase.MOVING else 0

                # The piece selected in the previous step (1 for the selected
                # piece, 0 elsewhere), if applicable, otherwise all 0’s.
                # Shape: _NUM_COLS * _NUM_ROWS tensor -> _NUM_COLS * _NUM_ROWS * 1 tensor.
                _idx = 3 * _NUM_CHESS_TYPES + self.num_history_move + 4
                obs[row][col][_idx] = 1 if (state.game_phase == GamePhase.MOVING
                                            and state.selected_pos[player] == [row, col]) else 0

        # self.string_from(state, player) # Uncomment this if you are debugging on observations

    def string_from(self, state, player):
        """Observation of `state` from the PoV of `player`, as a string."""
        # TODO: Add string formed observation for debugging and logging.
        obs = self.dict["observation"]
        obs_title = ["Prv.", "Pub. -i", "Pub. i", "History", "len1",
                                "len2", "Game phase", "Select", "Prev.Selection"]
        str_obs = [""] * 9

        for row in range(_NUM_ROWS):
            for col in range(_NUM_COLS):
                str_obs[0] += ",".join(map(str, np.where(obs[row][col][0:_NUM_CHESS_TYPES] == 1)[0] + 1)) + ", "
                str_obs[1] += ",".join(map(str, [obs[row][col][_NUM_CHESS_TYPES]])) + ", "
                str_obs[2] += ",".join(map(str, [obs[row][col][2 * _NUM_CHESS_TYPES]])) + ", "
                str_obs[3] += ",".join(map(str, [obs[row][col][3 * _NUM_CHESS_TYPES + self.mov_idx]])) + ", "
                str_obs[4] += ",".join(map(str,
                                           [obs[row][col][3 * _NUM_CHESS_TYPES + self.num_history_move]])) + ", "
                str_obs[5] += ",".join(map(str,
                                           [obs[row][col][3 * _NUM_CHESS_TYPES + self.num_history_move + 1]])) + ", "
                str_obs[6] += ",".join(map(str,
                                           [obs[row][col][3 * _NUM_CHESS_TYPES + self.num_history_move + 2]])) + ", "
                str_obs[7] += ",".join(map(str,
                                           [obs[row][col][3 * _NUM_CHESS_TYPES + self.num_history_move + 3]])) + ", "
                str_obs[8] += ",".join(map(str,
                                           [obs[row][col][3 * _NUM_CHESS_TYPES + self.num_history_move + 4]])) + ", "

            for i in range(len(str_obs)):
                str_obs[i] += "\n"

        print("Current Player: " + str(player), str(state.current_player()))
        [print(obs_title[i] + "\n" + str_obs[i]) for i in range(len(obs_title))]


# Helper functions for game details.


def _board_to_string(board):
    """Returns a string representation of the board."""
    return "\n".join("".join([str(chess) for chess in row]) for row in board)

def _board_to_string_pov(pov_board, board):
    """Returns a string representation of the board."""
    str_board: str = ""
    for row in range(_NUM_ROWS):
        for col in range(_NUM_COLS):
            str_board += str(pov_board[row][col])
        str_board += "      |      "
        for col in range(_NUM_COLS):
            str_board += str(board[row][col])
        str_board += "\n"
    return str_board


class Chess:
    def __init__(self, num=-10, country=-1):
        self.num = num
        self.type = ChessType(num)
        self.name = _DICT_CHESS_NAME[num]
        self.country = country if num != 0 else -1

    def __str__(self):
        if self.type == ChessType.NONE:
            return f"\033[;;m{self.name}\033[0m"
        elif self.country == 0:
            return f"\033[;30;43m{self.name}\033[0m"
        elif self.country == 1:
            return f"\033[;30;42m{self.name}\033[0m"

    def __repr__(self):
        return repr(self.num)

    def __eq__(self, other):
        return self.num == other

    def __lt__(self, other):
        return self.num < other

    def __gt__(self, other):
        return self.num > other


# Register the game with the OpenSpiel library
pyspiel.register_game(_GAME_TYPE, JunQiGame)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Note: Conan is supported on a best-effort basis. Abseil doesn't use Conan
# internally, so we won't know if it stops working. We may ask community
# members to help us debug any problems that arise.

from conans import ConanFile, CMake, tools
from conans.errors import ConanInvalidConfiguration
from conans.model.version import Version


class AbseilConan(ConanFile):
    name = "abseil"
    url = "https://github.com/abseil/abseil-cpp"
    homepage = url
    author = "Abseil <abseil-io@googlegroups.com>"
    description = "Abseil Common Libraries (C++) from Google"
    license = "Apache-2.0"
    topics = ("conan", "abseil", "abseil-cpp", "google", "common-libraries")
    exports = ["LICENSE"]
    exports_sources = ["CMakeLists.txt", "CMake/*", "absl/*"]
    generators = "cmake"
    settings = "os", "arch", "compiler", "build_type"

    def configure(self):
        if self.settings.os == "Windows" and \
           self.settings.compiler == "Visual Studio" and \
           Version(self.settings.compiler.version.value) < "14":
            raise ConanInvalidConfiguration("Abseil does not support MSVC < 14")

    def build(self):
        tools.replace_in_file("CMakeLists.txt", "project(absl LANGUAGES CXX)", "project(absl LANGUAGES CXX)\ninclude(conanbuildinfo.cmake)\nconan_basic_setup()")
        cmake = CMake(self)
        cmake.definitions["BUILD_TESTING"] = False
        cmake.configure()
        cmake.build()

    def package(self):
        self.copy("LICENSE", dst="licenses")
        self.copy("*.h", dst="include", src=".")
        self.copy("*.inc", dst="include", src=".")
        self.copy("*.a", dst="lib", src=".", keep_path=False)
        self.copy("*.lib", dst="lib", src=".", keep_path=False)

    def package_info(self):
        if self.settings.os == "Linux":
            self.cpp_info.libs = ["-Wl,--start-group"]
        self.cpp_info.libs.extend(tools.collect_libs(self))
        if self.settings.os == "Linux":
            self.cpp_info.libs.extend(["-Wl,--end-group", "pthread"])



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as python3
"""Tests for Python Linear Quadratic game."""

from absl.testing import absltest
import numpy as np
from open_spiel.python.mfg.games import linear_quadratic
import pyspiel

MFG_STR_CONST = "_a"


class MFGLinearQuadraticGameTest(absltest.TestCase):

  def test_load(self):
    game = pyspiel.load_game("mean_field_lin_quad")
    game.new_initial_state()

  def test_create(self):
    """Checks we can create the game and clone states."""
    game = linear_quadratic.MFGLinearQuadraticGame()
    self.assertEqual(game.size, linear_quadratic._SIZE)
    self.assertEqual(game.horizon, linear_quadratic._HORIZON)
    self.assertEqual(game.get_type().dynamics,
                     pyspiel.GameType.Dynamics.MEAN_FIELD)
    print("Num distinct actions:", game.num_distinct_actions())
    state = game.new_initial_state()
    clone = state.clone()
    print("Initial state:", state)
    print("Cloned initial state:", clone)

  def test_create_with_params(self):
    game = pyspiel.load_game("mean_field_lin_quad(horizon=30,size=100)")
    self.assertEqual(game.size, 100)
    self.assertEqual(game.horizon, 30)

  def check_cloning(self, state):
    cloned = state.clone()
    self.assertEqual(str(cloned), str(state))
    self.assertEqual(cloned._distribution, state._distribution)
    self.assertEqual(cloned._returns(), state._returns())
    self.assertEqual(cloned.current_player(), state.current_player())
    self.assertEqual(cloned.size, state.size)
    self.assertEqual(cloned.horizon, state.horizon)
    self.assertEqual(cloned._last_action, state._last_action)

  def test_random_game(self):
    """Tests basic API functions."""
    np.random.seed(7)
    horizon = 30
    size = 100
    game = linear_quadratic.MFGLinearQuadraticGame(params={
        "horizon": horizon,
        "size": size
    })
    state = game.new_initial_state()
    t = 0
    while not state.is_terminal():
      if state.current_player() == pyspiel.PlayerId.CHANCE:
        actions, probs = zip(*state.chance_outcomes())
        action = np.random.choice(actions, p=probs)
        self.check_cloning(state)
        self.assertEqual(len(state.legal_actions()),
                         len(state.chance_outcomes()))
        state.apply_action(action)
      elif state.current_player() == pyspiel.PlayerId.MEAN_FIELD:
        self.assertEqual(state.legal_actions(), [])
        self.check_cloning(state)
        num_states = len(state.distribution_support())
        state.update_distribution([1 / num_states] * num_states)
      else:
        self.assertEqual(state.current_player(), 0)
        self.check_cloning(state)
        state.observation_string()
        state.information_state_string()
        legal_actions = state.legal_actions()
        action = np.random.choice(legal_actions)
        state.apply_action(action)
        t += 1

    self.assertEqual(t, horizon)


if __name__ == "__main__":
  absltest.main()



# Copyright 2022 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""An implementation of PPO.

Note: code adapted (with permission) from
https://github.com/vwxyzjn/cleanrl/blob/master/cleanrl/ppo.py and
https://github.com/vwxyzjn/ppo-implementation-details/blob/main/ppo_atari.py.

Currently only supports the single-agent case.
"""

import time

import numpy as np
import torch
from torch import nn
from torch import optim
from torch.distributions.categorical import Categorical

from open_spiel.python.rl_agent import StepOutput

INVALID_ACTION_PENALTY = -1e6


def layer_init(layer, std=np.sqrt(2), bias_const=0.0):
  torch.nn.init.orthogonal_(layer.weight, std)
  torch.nn.init.constant_(layer.bias, bias_const)
  return layer


class CategoricalMasked(Categorical):
  """A masked categorical."""

  # pylint: disable=dangerous-default-value
  def __init__(self,
               probs=None,
               logits=None,
               validate_args=None,
               masks=[],
               mask_value=None):
    logits = torch.where(masks.bool(), logits, mask_value)
    super(CategoricalMasked, self).__init__(probs, logits, validate_args)


class PPOAgent(nn.Module):
  """A PPO agent module."""

  def __init__(self, num_actions, observation_shape, device):
    super().__init__()
    self.critic = nn.Sequential(
        layer_init(nn.Linear(np.array(observation_shape).prod(), 64)),
        nn.Tanh(),
        layer_init(nn.Linear(64, 64)),
        nn.Tanh(),
        layer_init(nn.Linear(64, 1), std=1.0),
    )
    self.actor = nn.Sequential(
        layer_init(nn.Linear(np.array(observation_shape).prod(), 64)),
        nn.Tanh(),
        layer_init(nn.Linear(64, 64)),
        nn.Tanh(),
        layer_init(nn.Linear(64, num_actions), std=0.01),
    )
    self.device = device
    self.num_actions = num_actions
    self.register_buffer("mask_value", torch.tensor(INVALID_ACTION_PENALTY))

  def get_value(self, x):
    return self.critic(x)

  def get_action_and_value(self, x, legal_actions_mask=None, action=None):
    if legal_actions_mask is None:
      legal_actions_mask = torch.ones((len(x), self.num_actions)).bool()

    logits = self.actor(x)
    probs = CategoricalMasked(
        logits=logits, masks=legal_actions_mask, mask_value=self.mask_value)
    if action is None:
      action = probs.sample()
    return action, probs.log_prob(action), probs.entropy(), self.critic(
        x), probs.probs


class PPOAtariAgent(nn.Module):
  """A PPO Atari agent module."""

  def __init__(self, num_actions, observation_shape, device):
    super(PPOAtariAgent, self).__init__()
    # Note: this network is intended for atari games, taken from
    # https://github.com/vwxyzjn/ppo-implementation-details/blob/main/ppo_atari.py
    self.network = nn.Sequential(
        layer_init(nn.Conv2d(4, 32, 8, stride=4)),
        nn.ReLU(),
        layer_init(nn.Conv2d(32, 64, 4, stride=2)),
        nn.ReLU(),
        layer_init(nn.Conv2d(64, 64, 3, stride=1)),
        nn.ReLU(),
        nn.Flatten(),
        layer_init(nn.Linear(64 * 7 * 7, 512)),
        nn.ReLU(),
    )
    self.actor = layer_init(nn.Linear(512, num_actions), std=0.01)
    self.critic = layer_init(nn.Linear(512, 1), std=1)
    self.num_actions = num_actions
    self.device = device
    self.register_buffer("mask_value", torch.tensor(INVALID_ACTION_PENALTY))

  def get_value(self, x):
    return self.critic(self.network(x / 255.0))

  def get_action_and_value(self, x, legal_actions_mask=None, action=None):
    if legal_actions_mask is None:
      legal_actions_mask = torch.ones((len(x), self.num_actions)).bool()

    hidden = self.network(x / 255.0)
    logits = self.actor(hidden)
    probs = CategoricalMasked(
        logits=logits, masks=legal_actions_mask, mask_value=self.mask_value)

    if action is None:
      action = probs.sample()
    return action, probs.log_prob(action), probs.entropy(), self.critic(
        hidden), probs.probs


def legal_actions_to_mask(legal_actions_list, num_actions):
  """Converts a list of legal actions to a mask.

  The mask has size num actions with a 1 in a legal positions.

  Args:
    legal_actions_list: the list of legal actions
    num_actions: number of actions (width of mask)

  Returns:
    legal actions mask.
  """
  legal_actions_mask = torch.zeros((len(legal_actions_list), num_actions),
                                   dtype=torch.bool)
  for i, legal_actions in enumerate(legal_actions_list):
    legal_actions_mask[i, legal_actions] = 1
  return legal_actions_mask


class PPO(nn.Module):
  """PPO Agent implementation in PyTorch.

  See open_spiel/python/examples/ppo_example.py for an usage example.

  Note that PPO runs multiple environments concurrently on each step (see
  open_spiel/python/vector_env.py). In practice, this tends to improve PPO's
  performance. The number of parallel environments is controlled by the
  num_envs argument.
  """

  def __init__(
      self,
      input_shape,
      num_actions,
      num_players,
      player_id=0,
      num_envs=1,
      steps_per_batch=128,
      num_minibatches=4,
      update_epochs=4,
      learning_rate=2.5e-4,
      gae=True,
      gamma=0.99,
      gae_lambda=0.95,
      normalize_advantages=True,
      clip_coef=0.2,
      clip_vloss=True,
      entropy_coef=0.01,
      value_coef=0.5,
      max_grad_norm=0.5,
      target_kl=None,
      device="cpu",
      writer=None,  # Tensorboard SummaryWriter
      agent_fn=PPOAtariAgent,
  ):
    super().__init__()

    self.input_shape = input_shape
    self.num_actions = num_actions
    self.num_players = num_players
    self.player_id = player_id
    self.device = device

    # Training settings
    self.num_envs = num_envs
    self.steps_per_batch = steps_per_batch
    self.batch_size = self.num_envs * self.steps_per_batch
    self.num_minibatches = num_minibatches
    self.minibatch_size = self.batch_size // self.num_minibatches
    self.update_epochs = update_epochs
    self.learning_rate = learning_rate

    # Loss function
    self.gae = gae
    self.gamma = gamma
    self.gae_lambda = gae_lambda
    self.normalize_advantages = normalize_advantages
    self.clip_coef = clip_coef
    self.clip_vloss = clip_vloss
    self.entropy_coef = entropy_coef
    self.value_coef = value_coef
    self.max_grad_norm = max_grad_norm
    self.target_kl = target_kl

    # Logging
    self.writer = writer

    # Initialize networks
    self.network = agent_fn(self.num_actions, self.input_shape,
                            device).to(device)
    self.optimizer = optim.Adam(
        self.parameters(), lr=self.learning_rate, eps=1e-5)

    # Initialize training buffers
    self.legal_actions_mask = torch.zeros(
        (self.steps_per_batch, self.num_envs, self.num_actions),
        dtype=torch.bool).to(device)
    self.obs = torch.zeros((self.steps_per_batch, self.num_envs) +
                           self.input_shape).to(device)
    self.actions = torch.zeros((self.steps_per_batch, self.num_envs)).to(device)
    self.logprobs = torch.zeros(
        (self.steps_per_batch, self.num_envs)).to(device)
    self.rewards = torch.zeros((self.steps_per_batch, self.num_envs)).to(device)
    self.dones = torch.zeros((self.steps_per_batch, self.num_envs)).to(device)
    self.values = torch.zeros((self.steps_per_batch, self.num_envs)).to(device)

    # Initialize counters
    self.cur_batch_idx = 0
    self.total_steps_done = 0
    self.updates_done = 0
    self.start_time = time.time()

  def get_value(self, x):
    return self.network.get_value(x)

  def get_action_and_value(self, x, legal_actions_mask=None, action=None):
    return self.network.get_action_and_value(x, legal_actions_mask, action)

  def step(self, time_step, is_evaluation=False):
    if is_evaluation:
      with torch.no_grad():
        legal_actions_mask = legal_actions_to_mask([
            ts.observations["legal_actions"][self.player_id] for ts in time_step
        ], self.num_actions).to(self.device)
        obs = torch.Tensor(
            np.array([
                np.reshape(ts.observations["info_state"][self.player_id],
                           self.input_shape) for ts in time_step
            ])).to(self.device)
        action, _, _, value, probs = self.get_action_and_value(
            obs, legal_actions_mask=legal_actions_mask)
        return [
            StepOutput(action=a.item(), probs=p)
            for (a, p) in zip(action, probs)
        ]
    else:
      with torch.no_grad():
        # act
        obs = torch.Tensor(
            np.array([
                np.reshape(ts.observations["info_state"][self.player_id],
                           self.input_shape) for ts in time_step
            ])).to(self.device)
        legal_actions_mask = legal_actions_to_mask([
            ts.observations["legal_actions"][self.player_id] for ts in time_step
        ], self.num_actions).to(self.device)
        action, logprob, _, value, probs = self.get_action_and_value(
            obs, legal_actions_mask=legal_actions_mask)

        # store
        self.legal_actions_mask[self.cur_batch_idx] = legal_actions_mask
        self.obs[self.cur_batch_idx] = obs
        self.actions[self.cur_batch_idx] = action
        self.logprobs[self.cur_batch_idx] = logprob
        self.values[self.cur_batch_idx] = value.flatten()

        agent_output = [
            StepOutput(action=a.item(), probs=p)
            for (a, p) in zip(action, probs)
        ]
        return agent_output

  def post_step(self, reward, done):
    self.rewards[self.cur_batch_idx] = torch.tensor(reward).to(
        self.device).view(-1)
    self.dones[self.cur_batch_idx] = torch.tensor(done).to(self.device).view(-1)

    self.total_steps_done += self.num_envs
    self.cur_batch_idx += 1

  def learn(self, time_step):
    next_obs = torch.Tensor(
        np.array([
            np.reshape(ts.observations["info_state"][self.player_id],
                       self.input_shape) for ts in time_step
        ])).to(self.device)

    # bootstrap value if not done
    with torch.no_grad():
      next_value = self.get_value(next_obs).reshape(1, -1)
      if self.gae:
        advantages = torch.zeros_like(self.rewards).to(self.device)
        lastgaelam = 0
        for t in reversed(range(self.steps_per_batch)):
          nextvalues = next_value if t == self.steps_per_batch - 1 else self.values[
              t + 1]
          nextnonterminal = 1.0 - self.dones[t]
          delta = self.rewards[
              t] + self.gamma * nextvalues * nextnonterminal - self.values[t]
          advantages[
              t] = lastgaelam = delta + self.gamma * self.gae_lambda * nextnonterminal * lastgaelam
        returns = advantages + self.values
      else:
        returns = torch.zeros_like(self.rewards).to(self.device)
        for t in reversed(range(self.steps_per_batch)):
          next_return = next_value if t == self.steps_per_batch - 1 else returns[
              t + 1]
          nextnonterminal = 1.0 - self.dones[t]
          returns[
              t] = self.rewards[t] + self.gamma * nextnonterminal * next_return
        advantages = returns - self.values

    # flatten the batch
    b_legal_actions_mask = self.legal_actions_mask.reshape(
        (-1, self.num_actions))
    b_obs = self.obs.reshape((-1,) + self.input_shape)
    b_logprobs = self.logprobs.reshape(-1)
    b_actions = self.actions.reshape(-1)
    b_advantages = advantages.reshape(-1)
    b_returns = returns.reshape(-1)
    b_values = self.values.reshape(-1)

    # Optimizing the policy and value network
    b_inds = np.arange(self.batch_size)
    clipfracs = []
    for _ in range(self.update_epochs):
      np.random.shuffle(b_inds)
      for start in range(0, self.batch_size, self.minibatch_size):
        end = start + self.minibatch_size
        mb_inds = b_inds[start:end]

        _, newlogprob, entropy, newvalue, _ = self.get_action_and_value(
            b_obs[mb_inds],
            legal_actions_mask=b_legal_actions_mask[mb_inds],
            action=b_actions.long()[mb_inds])
        logratio = newlogprob - b_logprobs[mb_inds]
        ratio = logratio.exp()

        with torch.no_grad():
          # calculate approx_kl http://joschu.net/blog/kl-approx.html
          old_approx_kl = (-logratio).mean()
          approx_kl = ((ratio - 1) - logratio).mean()
          clipfracs += [
              ((ratio - 1.0).abs() > self.clip_coef).float().mean().item()
          ]

        mb_advantages = b_advantages[mb_inds]
        if self.normalize_advantages:
          mb_advantages = (mb_advantages - mb_advantages.mean()) / (
              mb_advantages.std() + 1e-8)

        # Policy loss
        pg_loss1 = -mb_advantages * ratio
        pg_loss2 = -mb_advantages * torch.clamp(ratio, 1 - self.clip_coef,
                                                1 + self.clip_coef)
        pg_loss = torch.max(pg_loss1, pg_loss2).mean()

        # Value loss
        newvalue = newvalue.view(-1)
        if self.clip_vloss:
          v_loss_unclipped = (newvalue - b_returns[mb_inds])**2
          v_clipped = b_values[mb_inds] + torch.clamp(
              newvalue - b_values[mb_inds],
              -self.clip_coef,
              self.clip_coef,
          )
          v_loss_clipped = (v_clipped - b_returns[mb_inds])**2
          v_loss_max = torch.max(v_loss_unclipped, v_loss_clipped)
          v_loss = 0.5 * v_loss_max.mean()
        else:
          v_loss = 0.5 * ((newvalue - b_returns[mb_inds])**2).mean()

        entropy_loss = entropy.mean()
        loss = pg_loss - self.entropy_coef * entropy_loss + v_loss * self.value_coef

        self.optimizer.zero_grad()
        loss.backward()
        nn.utils.clip_grad_norm_(self.parameters(), self.max_grad_norm)
        self.optimizer.step()

      if self.target_kl is not None:
        if approx_kl > self.target_kl:
          break

    y_pred, y_true = b_values.cpu().numpy(), b_returns.cpu().numpy()
    var_y = np.var(y_true)
    explained_var = np.nan if var_y == 0 else 1 - np.var(y_true -
                                                         y_pred) / var_y

    # TRY NOT TO MODIFY: record rewards for plotting purposes
    if self.writer is not None:
      self.writer.add_scalar("charts/learning_rate",
                             self.optimizer.param_groups[0]["lr"],
                             self.total_steps_done)
      self.writer.add_scalar("losses/value_loss", v_loss.item(),
                             self.total_steps_done)
      self.writer.add_scalar("losses/policy_loss", pg_loss.item(),
                             self.total_steps_done)
      self.writer.add_scalar("losses/entropy", entropy_loss.item(),
                             self.total_steps_done)
      self.writer.add_scalar("losses/old_approx_kl", old_approx_kl.item(),
                             self.total_steps_done)
      self.writer.add_scalar("losses/approx_kl", approx_kl.item(),
                             self.total_steps_done)
      self.writer.add_scalar("losses/clipfrac", np.mean(clipfracs),
                             self.total_steps_done)
      self.writer.add_scalar("losses/explained_variance", explained_var,
                             self.total_steps_done)
      self.writer.add_scalar(
          "charts/SPS",
          int(self.total_steps_done / (time.time() - self.start_time)),
          self.total_steps_done)

    # Update counters
    self.updates_done += 1
    self.cur_batch_idx = 0

  def anneal_learning_rate(self, update, num_total_updates):
    # Annealing the rate
    frac = 1.0 - (update / num_total_updates)
    if frac <= 0:
      raise ValueError("Annealing learning rate to <= 0")
    lrnow = frac * self.learning_rate
    self.optimizer.param_groups[0]["lr"] = lrnow



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from absl.testing import absltest
import torch
import torch.nn.functional as F

import pyspiel
from open_spiel.python.pytorch import neurd

_GAME = pyspiel.load_game('kuhn_poker')


def _new_model():
  return neurd.DeepNeurdModel(
      _GAME,
      num_hidden_layers=1,
      num_hidden_units=13,
      num_hidden_factors=1,
      use_skip_connections=True,
      autoencode=True)


class NeurdTest(absltest.TestCase):

  def setUp(self):
    super(NeurdTest, self).setUp()
    torch.manual_seed(42)

  def test_neurd(self):
    num_iterations = 2
    models = [_new_model() for _ in range(_GAME.num_players())]

    solver = neurd.CounterfactualNeurdSolver(_GAME, models)

    average_policy = solver.average_policy()
    self.assertGreater(pyspiel.nash_conv(_GAME, average_policy), 0.91)

    def _train(model, data):
      neurd.train(
          model=model,
          data=data,
          batch_size=12,
          step_size=10.0,
          autoencoder_loss=F.huber_loss)

    for _ in range(num_iterations):
      solver.evaluate_and_update_policy(_train)

    average_policy = solver.average_policy()
    self.assertLess(pyspiel.nash_conv(_GAME, average_policy), 0.91)


if __name__ == '__main__':
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from absl.testing import absltest

import numpy as np
from open_spiel.python.algorithms import nfg_utils


class NfgUtilsTest(absltest.TestCase):

  def test_strategy_averager_len_smaller_than_window(self):
    averager = nfg_utils.StrategyAverager(2, [2, 2], window_size=50)
    averager.append([np.array([1.0, 0.0]), np.array([0.0, 1.0])])
    averager.append([np.array([0.0, 1.0]), np.array([1.0, 0.0])])
    avg_strategies = averager.average_strategies()
    self.assertLen(avg_strategies, 2)
    self.assertAlmostEqual(avg_strategies[0][0], 0.5)
    self.assertAlmostEqual(avg_strategies[0][1], 0.5)
    self.assertAlmostEqual(avg_strategies[1][0], 0.5)
    self.assertAlmostEqual(avg_strategies[1][1], 0.5)

  def test_strategy_averager(self):
    first_action_strat = np.array([1.0, 0.0])
    second_action_strat = np.array([0.0, 1.0])
    averager_full = nfg_utils.StrategyAverager(2, [2, 2])
    averager_window5 = nfg_utils.StrategyAverager(2, [2, 2], window_size=5)
    averager_window6 = nfg_utils.StrategyAverager(2, [2, 2], window_size=6)
    for _ in range(5):
      averager_full.append([first_action_strat, first_action_strat])
      averager_window5.append([first_action_strat, first_action_strat])
      averager_window6.append([first_action_strat, first_action_strat])
    for _ in range(5):
      averager_full.append([second_action_strat, second_action_strat])
      averager_window5.append([second_action_strat, second_action_strat])
      averager_window6.append([second_action_strat, second_action_strat])
    avg_full = averager_full.average_strategies()
    avg_window5 = averager_window5.average_strategies()
    avg_window6 = averager_window6.average_strategies()
    self.assertAlmostEqual(avg_full[0][1], 0.5)
    self.assertAlmostEqual(avg_window5[0][1], 5.0 / 5.0)
    self.assertAlmostEqual(avg_window6[0][1], 5.0 / 6.0)


if __name__ == '__main__':
  absltest.main()




# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as python3
"""Tests for Python Crowd avoidance game."""

from absl.testing import absltest
from absl.testing import parameterized
import numpy as np
import numpy.testing as npt
from open_spiel.python.mfg.games import crowd_avoidance
import pyspiel


class MFGCrowdAvoidanceGameTest(parameterized.TestCase):

  def test_load(self):
    game = pyspiel.load_game('python_mfg_crowd_avoidance')
    game.new_initial_state_for_population(0)
    game.new_initial_state_for_population(1)

  @parameterized.parameters(
      {
          'geometry': crowd_avoidance.Geometry.SQUARE,
          'expected_pos': np.array([5, 3]),
      },
      {
          'geometry': crowd_avoidance.Geometry.TORUS,
          'expected_pos': np.array([5, 3]),
      },
  )
  def test_dynamics(self, geometry, expected_pos):
    game = pyspiel.load_game(
        'python_mfg_crowd_avoidance',
        {
            'geometry': geometry,
        },
    )
    state = game.new_initial_state_for_population(1)
    # Initial chance node.
    self.assertEqual(state.current_player(), pyspiel.PlayerId.CHANCE)
    self.assertLen(state.chance_outcomes(), 3)
    self.assertEqual(
        state.chance_outcomes()[0][0],
        crowd_avoidance.pos_to_merged(np.array([5, 2]), state.size),
    )
    state.apply_action(state.chance_outcomes()[0][0])
    self.assertEqual(state.current_player(), 1)
    npt.assert_array_equal(state.pos, [5, 2])
    self.assertEqual(state._action_to_string(player=1, action=2), '[0 1]')
    state.apply_action(2)
    npt.assert_array_equal(state.pos, expected_pos)

  def test_create_with_params(self):
    setting = 'python_mfg_crowd_avoidance()'
    game = pyspiel.load_game(setting)
    self.assertEqual(game.size, 7)
    self.assertEqual(game.horizon, 10)

  @parameterized.parameters(
      {'population': 0},
      {'population': 1},
  )
  def test_random_game(self, population):
    """Tests basic API functions."""
    congestion_matrix = np.array([[0, 1], [1, 0]])
    init_distrib = np.array([
        # First population
        [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
        [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
        [0.0, 0.4, 0.4, 0.0, 0.0, 0.0, 0.0],
        [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
        [0.0, 0.2, 0.0, 0.0, 0.0, 0.0, 0.0],
        [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
        [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
        # Second population
        [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
        [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
        [0.0, 0.0, 0.0, 0.0, 0.0, 0.2, 0.0],
        [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
        [0.0, 0.0, 0.0, 0.0, 0.4, 0.4, 0.0],
        [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
        [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
    ])
    forbidden_states_grid = [
        '#######',
        '#  #  #',
        '#     #',
        '#  #  #',
        '#     #',
        '#  #  #',
        '#######',
    ]
    forbidden_states = crowd_avoidance.grid_to_forbidden_states(
        forbidden_states_grid
    )
    game = crowd_avoidance.MFGCrowdAvoidanceGame(
        params={
            'horizon': 10,
            'size': 7,
            'players': 2,
            'congestion_matrix': ' '.join(
                str(v) for v in congestion_matrix.flatten()
            ),
            'init_distrib': ' '.join(str(v) for v in init_distrib.flatten()),
            'forbidden_states': forbidden_states,
        }
    )
    pyspiel.random_sim_test(
        game,
        num_sims=10,
        serialize=False,
        verbose=True,
        mean_field_population=population,
    )

  @parameterized.parameters(
      {
          'coef_congestion': 1.5,
          'coef_target': 0.6,
          'congestion_matrix': np.array([[0, 1], [1, 0]]),
          'population': 0,
          'players': 2,
          'initial_pos': np.array([0, 0]),
          'distributions': [
              # First population
              np.array([[0.8, 0.2], [0.0, 0.0]]),
              # Second population
              np.array([[0.3, 0.7], [0.0, 0.0]]),
          ],
          'expected_rewards': np.array([
              -1.5 * 0.3 + 0.0,
              -1.5 * 0.8 + 0.0,
          ]),
          'init_distrib': np.array([
              # First population
              [0.8, 0.2],
              [0.0, 0.0],
              # Second population
              [0.3, 0.7],
              [0.0, 0.0],
          ]),
      },
  )
  def test_rewards(
      self,
      coef_congestion,
      coef_target,
      congestion_matrix,
      players,
      population,
      initial_pos,
      distributions,
      expected_rewards,
      init_distrib,
  ):
    game = pyspiel.load_game(
        'python_mfg_crowd_avoidance',
        {
            'size': 2,
            'coef_congestion': coef_congestion,
            'coef_target': coef_target,
            'congestion_matrix': ' '.join(
                str(v) for v in congestion_matrix.flatten()
            ),
            'players': players,
            'init_distrib': ' '.join(str(v) for v in init_distrib.flatten()),
            'forbidden_states': '[]',
        },
    )
    state = game.new_initial_state_for_population(population)
    # Initial chance node.
    self.assertEqual(state.current_player(), pyspiel.PlayerId.CHANCE)
    state.apply_action(crowd_avoidance.pos_to_merged(initial_pos, state.size))
    self.assertEqual(state.current_player(), population)
    npt.assert_array_equal(state.pos, initial_pos)
    state.apply_action(state._NEUTRAL_ACTION)
    npt.assert_array_equal(state.pos, initial_pos)
    self.assertEqual(state.current_player(), pyspiel.PlayerId.CHANCE)
    state.apply_action(state._NEUTRAL_ACTION)
    self.assertEqual(state.current_player(), pyspiel.PlayerId.MEAN_FIELD)

    # Maps states (in string representation) to their proba.
    dist = {}
    for x in range(state.size):
      for y in range(state.size):
        for pop in range(len(congestion_matrix)):
          state_str = state.state_to_str(
              np.array([x, y]),
              state.t,
              pop,
              player_id=pyspiel.PlayerId.MEAN_FIELD,
          )
          dist[state_str] = distributions[pop][y][x]
    support = state.distribution_support()
    state.update_distribution([dist[s] for s in support])

    # Decision node where we get a reward.
    self.assertEqual(state.current_player(), population)
    npt.assert_array_equal(state.rewards(), expected_rewards)


if __name__ == '__main__':
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for open_spiel.python.algorithms.policy_aggregator_joint."""

from absl.testing import absltest
from absl.testing import parameterized

import numpy as np

from open_spiel.python import policy
from open_spiel.python import rl_environment
from open_spiel.python.algorithms import policy_aggregator_joint


class JointPolicyAggregatorTest(parameterized.TestCase):

  @parameterized.named_parameters(
      {
          "testcase_name": "kuhn_poker",
          "game_name": "kuhn_poker"
      }, {
          "testcase_name": "leduc_poker",
          "game_name": "leduc_poker"
      })
  def test_policy_aggregation_random(self, game_name):
    env = rl_environment.Environment(game_name)
    num_players = 2
    num_joint_policies = 4

    joint_policies = [[
        policy.UniformRandomPolicy(env.game) for _ in range(num_players)
    ] for _ in range(num_joint_policies)]
    probabilities = np.ones(len(joint_policies))
    probabilities /= np.sum(probabilities)

    pol_ag = policy_aggregator_joint.JointPolicyAggregator(env.game)
    aggr_policy = pol_ag.aggregate([0, 1], joint_policies, probabilities)

    self.assertLen(aggr_policy.policies, num_players)
    for player in range(num_players):
      player_policy = aggr_policy.policies[player]
      self.assertNotEmpty(player_policy)
      for state_action_probs in player_policy.values():
        probs = list(state_action_probs.values())
        expected_prob = 1. / len(probs)
        for prob in probs:
          self.assertAlmostEqual(expected_prob, prob, places=10)


if __name__ == "__main__":
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for open_spiel.python.algorithms.eva."""

from absl.testing import absltest
from absl.testing import parameterized
import torch

from open_spiel.python import rl_environment
from open_spiel.python.pytorch import eva


SEED = 24984617


class EVATest(parameterized.TestCase):

  @parameterized.parameters("tic_tac_toe", "kuhn_poker", "liars_dice")
  def test_run_games(self, game):
    env = rl_environment.Environment(game)
    num_players = env.num_players
    eva_agents = []
    num_actions = env.action_spec()["num_actions"]
    state_size = env.observation_spec()["info_state"][0]
    for player in range(num_players):
      eva_agents.append(
          eva.EVAAgent(
              env,
              player,
              state_size,
              num_actions,
              embedding_network_layers=(64, 32),
              embedding_size=12,
              learning_rate=1e-4,
              mixing_parameter=0.5,
              memory_capacity=int(1e6),
              discount_factor=1.0,
              epsilon_start=1.0,
              epsilon_end=0.1,
              epsilon_decay_duration=int(1e6)))
    time_step = env.reset()
    while not time_step.last():
      current_player = time_step.observations["current_player"]
      current_agent = eva_agents[current_player]
      # 1.  Step the agent.
      # 2.  Step the Environment.
      agent_output = current_agent.step(time_step)
      time_step = env.step([agent_output.action])
    for agent in eva_agents:
      agent.step(time_step)


class QueryableFixedSizeRingBufferTest(absltest.TestCase):

  def test_replay_buffer_add(self):
    replay_buffer = eva.QueryableFixedSizeRingBuffer(replay_buffer_capacity=10)
    self.assertEmpty(replay_buffer)
    replay_buffer.add("entry1")
    self.assertLen(replay_buffer, 1)
    replay_buffer.add("entry2")
    self.assertLen(replay_buffer, 2)

    self.assertIn("entry1", replay_buffer)
    self.assertIn("entry2", replay_buffer)

  def test_replay_buffer_max_capacity(self):
    replay_buffer = eva.QueryableFixedSizeRingBuffer(replay_buffer_capacity=2)
    replay_buffer.add("entry1")
    replay_buffer.add("entry2")
    replay_buffer.add("entry3")
    self.assertLen(replay_buffer, 2)

    self.assertIn("entry2", replay_buffer)
    self.assertIn("entry3", replay_buffer)

  def test_replay_buffer_sample(self):
    replay_buffer = eva.QueryableFixedSizeRingBuffer(replay_buffer_capacity=3)
    replay_buffer.add("entry1")
    replay_buffer.add("entry2")
    replay_buffer.add("entry3")

    samples = replay_buffer.sample(3)

    self.assertIn("entry1", samples)
    self.assertIn("entry2", samples)
    self.assertIn("entry3", samples)

  # TODO(author6) Test knn query.


if __name__ == "__main__":
  torch.manual_seed(SEED)
  absltest.main()



"""Tests that the casters raise a TypeError if the status module is missing."""

from absl.testing import absltest
from pybind11_abseil.tests import missing_import


class MissingStatusImportTest(absltest.TestCase):

  message_regex = 'Status module has not been imported.*'

  def test_returns_status(self):
    with self.assertRaisesRegex(TypeError, self.message_regex):
      missing_import.returns_status()

  def test_returns_status_or(self):
    with self.assertRaisesRegex(TypeError, self.message_regex):
      missing_import.returns_status_or()


if __name__ == '__main__':
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests if a bot fails after a few actions.

A bot that picks the first action from the list for the first two rounds,
and then exists with an exception.
Used only for tests.
"""

import sys

game_name = input()
play_as = int(input())
print("ready")

while True:
  print("start")
  num_actions = 0
  while True:
    message = input()
    if message == "tournament over":
      print("tournament over")
      sys.exit(0)
    if message.startswith("match over"):
      print("match over")
      break
    public_buf, private_buf, *legal_actions = message.split(" ")
    if legal_actions:
      num_actions += 1
      print(legal_actions[-1])
    else:
      print("ponder")

    if num_actions > 2:
      raise RuntimeError



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Python RNAD example."""
import numpy
from absl import app
from absl import flags
from absl import logging
import matplotlib
import matplotlib.pyplot as plt
import time, sys, os
import numpy as np
import pickle
import time
import threading
import warnings

from open_spiel.python.algorithms.rnad import rnad_for_junqi as rnad

_exit = 0
_TIME_GAP = 60

FLAGS = flags.FLAGS

flags.DEFINE_integer("num_iterations", 100, "Number of iterations")
flags.DEFINE_integer("num_traversals", 150, "Number of traversals/games")
flags.DEFINE_string("game_name", "junqi1", "Name of the game")

loss_values = []
iterations_list = []


def command_line_action(time_step):
    """Gets a valid action from the user on the command line."""
    current_player = time_step.observations["current_player"]
    legal_actions = time_step.observations["legal_actions"][current_player]
    action = -1
    while action not in legal_actions:
        print("Choose an action from {}:".format(legal_actions))
        sys.stdout.flush()
        action_str = input()
        try:
            action = int(action_str)
        except ValueError:
            continue
    return action


def eval_against_random_bots(env, trained_agents, random_agents, num_episodes):
    """Evaluates `trained_agents` against `random_agents` for `num_episodes`."""
    wins = np.zeros(2)
    for player_pos in range(2):
        if player_pos == 0:
            cur_agents = [trained_agents[0], random_agents[1]]
        else:
            cur_agents = [random_agents[0], trained_agents[1]]
        for _ in range(num_episodes):
            time_step = env.reset()
            while not time_step.last():
                player_id = time_step.observations["current_player"]
                agent_output = cur_agents[player_id].step(time_step, is_evaluation=True)
                time_step = env.step([agent_output.action])
            if time_step.rewards[player_pos] > 0:
                wins[player_pos] += 1
    return wins / num_episodes


class JunQiSolver(rnad.RNaDSolver):
    pass


class HiddenPrints:
    def __enter__(self):
        self._original_stdout = sys.stdout
        sys.stdout = open(os.devnull, 'w')

    def __exit__(self, exc_type, exc_val, exc_tb):
        sys.stdout.close()
        sys.stdout = self._original_stdout


def plot():
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        plt.ion()  # Turn on interactive mode for dynamic updating
        fig, ax = plt.subplots()
        # matplotlib.use('TkAgg')
        ax.set_title('Training Loss Curve')
        ax.set_xlabel('Iteration')
        ax.set_ylabel('Loss')
        while not _exit:
            plt.pause(_TIME_GAP)
            ax.clear()
            ax.plot(iterations_list, loss_values)  # , marker='o', linestyle='-')
            ax.set_title('Training Loss Curve')
            ax.set_xlabel('Iteration')
            ax.set_ylabel('Loss')
            plt.pause(0.01)
            plt.draw()
            plt.savefig(str(len(iterations_list)))
            plt.pause(0.01)


def print_loss(any, i):
    loss_values.append(any['loss'].tolist())
    iterations_list.append(any['learner_steps'])
    if i % 1 == 0:
        print(
            f"[DATA] Loss: {round(any['loss'].tolist(), 4)}, "
              f"Actor Steps: {any['actor_steps']}, "
              f"Learner Steps: {any['learner_steps']}"
        )


config = rnad.RNaDConfig(
    game_name='junqi1',
    trajectory_max=200,
    state_representation=rnad.StateRepresentation.OBSERVATION,
    policy_network_layers=(256, 256),
    batch_size=1,
    learning_rate=0.00005,
    adam=rnad.AdamConfig(),
    clip_gradient=10_000,
    target_network_avg=0.001,
    entropy_schedule_repeats=(1,),
    entropy_schedule_size=(20_000,),
    eta_reward_transform=0.2,
    nerd=rnad.NerdConfig(),
    c_vtrace=1.0,
    finetune=rnad.FineTuning(),
    seed=42,
)


def main(unused_argv):
    # logging.info("Loading %s", FLAGS.game_name)
    # game = pyspiel.load_game(FLAGS.game_name)
    with open('model.pkl', 'wb') as f:
        pass
    with open('model.pkl', 'rb') as f:
        if f.readline() != b'':
            print("[INFO] Model data found, continuing training...")
            rnad_solver = pickle.load(f)
        else:
            rnad_solver = JunQiSolver(config)
    i = 0
    iterations = 1e2
    t_list = []
    t_std = time.perf_counter()
    threading.Thread(target=plot).start()
    try:
        while (i <= iterations):
            t_start = time.perf_counter()
            print_loss(rnad_solver.step(), i)
            t_end = time.perf_counter()
            t_list.append(t_end - t_start)
            print(
                f"[INFO] Training in progress: {100 * i / iterations}% [{i} of {iterations} iterations] " + "Time used: " + time.strftime(
                    "%H:%M:%S", time.gmtime(t_end - t_std)) + " ETA: " + time.strftime(
                    "%H:%M:%S", time.gmtime(numpy.average(t_list) * (iterations - i))))
            i += 1
    except KeyboardInterrupt:
        with open('model.pkl', 'wb') as f:
            pickle.dump(rnad_solver, f)
    with open('model.pkl', 'wb') as f:
        pickle.dump(rnad_solver, f)

    rnad_solver.pit()


if __name__ == "__main__":
    app.run(main)
    _exit = 1
    sys.exit()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Regret Matching Approximate Nash Solver."""

from absl import logging  # pylint:disable=unused-import

import numpy as np


class Solver(object):
  """Regret-matching Solver."""

  def __init__(self, optimism=True, discount=False, rnd_init=False, seed=None,
               **kwargs):
    """Ctor."""
    del kwargs
    self.num_players = None
    self.lrs = None
    self.optimism = optimism
    self.discount = discount
    self.rnd_init = rnd_init
    self.has_aux = True
    self.aux_errors = []

    self.seed = seed
    self.random = np.random.RandomState(seed)

  def init_vars(self, num_strats, num_players):
    """Initialize solver parameters."""
    self.num_players = num_players
    if len(num_strats) != num_players:
      raise ValueError('Must specify num strategies for each player')
    init_dist = []
    for num_strats_i in num_strats:
      if self.rnd_init:
        init_dist_i = self.random.rand(num_strats_i)
      else:
        init_dist_i = np.ones(num_strats_i)
      init_dist_i /= init_dist_i.sum()
      init_dist.append(init_dist_i)
    init_regret = [np.zeros_like(dist_i) for dist_i in init_dist]
    return (init_dist, init_regret)

  def record_aux_errors(self, grads):
    """Record errors for the auxiliary variables."""
    grad_regret = grads[1]
    grad_regret_flat = np.concatenate(grad_regret)
    self.aux_errors.append([np.linalg.norm(grad_regret_flat)])

  def compute_gradients(self, params, payoff_matrices):
    """Compute and return gradients (and exploitabilities) for all parameters.

    Args:
      params: tuple of params (dist, regret), see regmatch.gradients
      payoff_matrices: dictionary with keys as tuples of agents (i, j) and
        values of (2 x A x A) np.arrays, payoffs for each joint action. keys
        are sorted and arrays should be indexed in the same order
    Returns:
      tuple of gradients (grad_dist, grad_regret), see ate.gradients
      unregularized exploitability (stochastic estimate)
      solver exploitability (stochastic estimate) - NaN
    """
    return gradients(*params, payoff_matrices, self.num_players)

  def exploitability(self, params, payoff_matrices):
    """Regret matching does not minimize any exploitability so return NaN.

    Args:
      params: tuple of params (dist,)
      payoff_matrices: dictionary with keys as tuples of agents (i, j) and
        values of (2 x A x A) np.arrays, payoffs for each joint action. keys
        are sorted and arrays should be indexed in the same order
    Returns:
      np.NaN
    """
    del params
    del payoff_matrices
    return np.NaN

  def update(self, params, grads, t):
    """Update cumulative regret and strategy (dist).

    Args:
      params: tuple of variables to be updated (dist, regret)
      grads: tuple of variable gradients (grad_dist, grad_regret)
      t: int, solver iteration (not used)
    Returns:
      new_params: tuple of update params (new_dist, new_regret)
    """
    dist, regret = params
    regret_delta = grads[1]
    if self.discount:
      gamma = t / float(t + 1)
    else:
      gamma = 1

    new_dist = []
    new_regret = []
    for dist_i, regret_i, regret_delta_i in zip(dist, regret, regret_delta):
      new_regret_i = gamma * regret_i + regret_delta_i
      new_clipped_regrets_i = np.clip(
          new_regret_i + self.optimism * regret_delta_i, 0., np.inf)
      if np.sum(new_clipped_regrets_i) > 0:
        new_dist_i = new_clipped_regrets_i / new_clipped_regrets_i.sum()
      else:
        new_dist_i = np.ones_like(dist_i) / dist_i.size
      new_dist.append(new_dist_i)
      new_regret.append(new_regret_i)

    new_params = (new_dist, new_regret)
    return new_params


def gradients(dist, regret, payoff_matrices, num_players):
  """Computes regret delta to be added to regret in update.

  Args:
    dist: list of 1-d np.arrays, current estimate of nash distribution
    regret: list of 1-d np.arrays (same as dist), current estimate of regrets
    payoff_matrices: dictionary with keys as tuples of agents (i, j) and
        values of (2 x A x A) np.arrays, payoffs for each joint action. keys
        are sorted and arrays should be indexed in the same order
    num_players: int, number of players, in case payoff_matrices is abbreviated
  Returns:
    deltas w.r.t. (dist, regret) as tuple
    unregularized exploitability (stochastic estimate)
    solver exploitability (stochastic estimate) - NaN
  """
  del regret

  # first compute best responses and payoff gradients
  grad_dist = []
  grad_regret = []
  unreg_exp = []
  for i in range(num_players):

    nabla_i = np.zeros_like(dist[i])
    # TODO(imgemp): decide if averaging over nablas provides best comparison
    for j in range(num_players):
      if j == i:
        continue
      if i < j:
        hess_i_ij = payoff_matrices[(i, j)][0]
      else:
        hess_i_ij = payoff_matrices[(j, i)][1].T

      nabla_ij = hess_i_ij.dot(dist[j])
      nabla_i += nabla_ij / float(num_players - 1)

    grad_dist_i = np.NaN * np.ones_like(nabla_i)
    grad_dist.append(grad_dist_i)

    utility_i = nabla_i.dot(dist[i])
    grad_regret_i = nabla_i - utility_i
    grad_regret.append(grad_regret_i)

    unreg_exp.append(np.max(nabla_i) - nabla_i.dot(dist[i]))

  return (grad_dist, grad_regret), np.mean(unreg_exp), np.NaN



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as python3
"""Tests for Python Block Dominoes."""

from absl.testing import absltest
from open_spiel.python.games import block_dominoes
import pyspiel


class DominoesBlockTest(absltest.TestCase):

  def test_game_from_cc(self):
    """Runs our standard game tests, checking API consistency."""
    game = pyspiel.load_game("python_block_dominoes")
    pyspiel.random_sim_test(game, num_sims=100, serialize=False, verbose=True)

  def test_single_deterministic_game_1(self):
    """Runs a single game where tiles and actions chose deterministically."""
    game = pyspiel.load_game("python_block_dominoes")
    state = game.new_initial_state()
    hand0 = [
        (6.0, 6.0),
        (0.0, 2.0),
        (4.0, 4.0),
        (3.0, 3.0),
        (2.0, 2.0),
        (1.0, 1.0),
        (0.0, 0.0),
    ]
    hand1 = [
        (5.0, 6.0),
        (4.0, 5.0),
        (3.0, 4.0),
        (2.0, 3.0),
        (1.0, 2.0),
        (0.0, 1.0),
        (4.0, 6.0),
    ]
    self.deal_hands(state, [hand0, hand1])

    self.apply_action(state, block_dominoes.Action(0, (6.0, 6.0), None))
    self.apply_action(state, block_dominoes.Action(1, (5.0, 6.0), 6.0))
    # player 0 don't hold any tile with 6 or 5, player 1 turn again
    self.apply_action(state, block_dominoes.Action(1, (4.0, 5.0), 5.0))
    self.apply_action(state, block_dominoes.Action(0, (4.0, 4.0), 4.0))
    self.apply_action(state, block_dominoes.Action(1, (3.0, 4.0), 4.0))
    self.apply_action(state, block_dominoes.Action(0, (3.0, 3.0), 3.0))
    self.apply_action(state, block_dominoes.Action(1, (2.0, 3.0), 3.0))
    self.apply_action(state, block_dominoes.Action(0, (2.0, 2.0), 2.0))
    self.apply_action(state, block_dominoes.Action(1, (1.0, 2.0), 2.0))
    self.apply_action(state, block_dominoes.Action(0, (1.0, 1.0), 1.0))
    self.apply_action(state, block_dominoes.Action(1, (0.0, 1.0), 1.0))
    self.apply_action(state, block_dominoes.Action(0, (0.0, 0.0), 0.0))
    self.apply_action(state, block_dominoes.Action(1, (4.0, 6.0), 6.0))

    # player 1 played all is tile and player 0 hold the tile (0, 2)
    self.assertTrue(state.is_terminal())
    self.assertEqual(state.returns()[0], -2)
    self.assertEqual(state.returns()[1], 2)

  def test_single_deterministic_game_2(self):
    """Runs a single game where tiles and actions chose deterministically."""
    game = pyspiel.load_game("python_block_dominoes")
    state = game.new_initial_state()
    hand0 = [
        (6.0, 6.0),
        (0.0, 5.0),
        (1.0, 5.0),
        (2.0, 5.0),
        (3.0, 5.0),
        (4.0, 5.0),
        (5.0, 5.0),
    ]
    hand1 = [
        (0.0, 4.0),
        (1.0, 4.0),
        (2.0, 4.0),
        (3.0, 4.0),
        (4.0, 4.0),
        (0.0, 3.0),
        (1.0, 3.0),
    ]
    self.deal_hands(state, [hand0, hand1])

    self.apply_action(state, block_dominoes.Action(0, (6.0, 6.0), None))
    # Both players don't hold tile with 6, therefore both blocked and the
    # game end
    self.assertTrue(state.is_terminal())
    self.assertEqual(state.returns()[0], -45)
    self.assertEqual(state.returns()[1], 45)

  @staticmethod
  def apply_action(state, action):
    actions_str = block_dominoes._ACTIONS_STR
    state.apply_action(actions_str.index(str(action)))

  @staticmethod
  def deal_hands(state, hands):
    deck = block_dominoes._DECK
    for hand in hands:
      for t in hand:
        state.apply_action(deck.index(t))


if __name__ == "__main__":
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as python3
"""Implementation of dynamic routing game.

The game is derived from https://arxiv.org/abs/2110.11943.
This dynamic routing game models the evolution of N vehicles in a road network.
The vehicles are described by their current link location, the time they have to
spend on the link before exiting it, and their destination. The action of a
vehicle is the successor link they want to reach when exiting a given link.
Actions are encoded as integer from 0 to K. Action 0 encodes not being able to
move on a successor link because the waiting time of the player is still
positive. Actions 1 to K correspond to the indices of the network links. Legal
actions for a player on link l, with a negative waiting time are the indices of
the successors link of l. When arriving on a link, the waiting time of the
player is assign based on the number of players on the link at this time. Over
time steps, the waiting time linearly decrease until it is negative, the vehicle
moves to a successor link and the waiting time get reassigned.
The cost of the vehicle is its arrival time, it could be seen as a running cost
where +1 is added to the cost at any time step the vehicle is not on its
destination.
This dynamic routing game is a mesoscopic traffic model with explicit congestion
dynamics where vehicle minimizes their arrival time.

The game is defined by:
- a network given by the class Network.
- a list of vehicles given by the class Vehicle.

The current game is implementated as a N player game. However this game can also
be extended to a mean field game, implemented as python_mfg_dynamic_routing.
"""

from typing import Any, Iterable, List, Mapping, Optional, Set

import numpy as np
from open_spiel.python.games import dynamic_routing_data
from open_spiel.python.games import dynamic_routing_utils
from open_spiel.python.observation import IIGObserverForPublicInfoGame
import pyspiel

_DEFAULT_PARAMS = {
    "max_num_time_step": 10,
    "time_step_length": 0.5,
    "players": -1
}
_GAME_TYPE = pyspiel.GameType(
    short_name="python_dynamic_routing",
    long_name="Python Dynamic Routing Game",
    dynamics=pyspiel.GameType.Dynamics.SIMULTANEOUS,
    chance_mode=pyspiel.GameType.ChanceMode.DETERMINISTIC,
    information=pyspiel.GameType.Information.PERFECT_INFORMATION,
    utility=pyspiel.GameType.Utility.GENERAL_SUM,
    reward_model=pyspiel.GameType.RewardModel.REWARDS,
    max_num_players=100,
    min_num_players=0,
    provides_information_state_string=True,
    provides_information_state_tensor=True,
    provides_observation_string=True,
    provides_observation_tensor=True,
    default_loadable=True,
    provides_factored_observation_string=True,
    parameter_specification=_DEFAULT_PARAMS)


class DynamicRoutingGame(pyspiel.Game):
  """Implementation of dynamic routing game.

  At each simultaneous-move time, each vehicle/player with negative waiting time
  chooses on which successor link they would like to go. When arriving on the
  link, a waiting time is assigned to the player based on the count of players
  on the link, after everyone has moved to their successors link. One vehicle
  arrival time is equal to the time step when they first reach their
  destination.
  See module docstring for more information.

  Attributes inherited from GameInfo:
    max_chance_outcome: 0, the game is deterministic.
    max_game_length: maximum number of time step played. Passed during
      construction.
    max_utility: maximum utility is the opposite of the minimum arrival time.
      Set to 0.
    min_utility: minimum utility is the opposite of the maximum arrival time.
      Set to - max_game_length - 1.
    num_distinct_actions: maximum number of possible actions. This is equal to
      the number of links + 1 (corresponding to having no possible action
      _NO_POSSIBLE_ACTION).
    num_players: the number of vehicles. Choosen during by the constructor as
      the number of vehicles.
  Attributes:
    network: the network of the game.
    _vehicles: a list of the vehicle. Their origin and their destination should
      be road sections of the game. The number of vehicles in the list sets the
      num_players attribute.
    time_step_length: size of the time step, used to convert travel times into
      number of game time steps.
    perform_sanity_checks: if true, sanity checks are done during the game,
      should be set to false to speed up the game.
  """
  network: dynamic_routing_utils.Network
  _vehicles: List[dynamic_routing_utils.Vehicle]
  perform_sanity_checks: bool
  time_step_length: float

  def __init__(
      self,
      params: Mapping[str, Any],
      network: Optional[dynamic_routing_utils.Network] = None,
      vehicles: Optional[List[dynamic_routing_utils.Vehicle]] = None,
      perform_sanity_checks: bool = True,
  ):
    """Initiliaze the game.

    Args:
      params: game parameters. It should define max_num_time_step and
        time_step_length.
      network: the network of the game.
      vehicles: a list of the vehicle. Their origin and their destination should
        be road sections of the game. The number of vehicles in the list sets
        the num_players attribute.
      perform_sanity_checks: set the perform_sanity_checks attribute.
    """
    max_num_time_step = params["max_num_time_step"]
    time_step_length = params["time_step_length"]
    self.network = network if network else dynamic_routing_data.BRAESS_NETWORK
    self._vehicles = (
        vehicles
        if vehicles else dynamic_routing_data.BRAESS_NETWORK_VEHICLES_DEMAND)
    self.network.check_list_of_vehicles_is_correct(self._vehicles)
    self.perform_sanity_checks = perform_sanity_checks
    self.time_step_length = time_step_length
    game_info = pyspiel.GameInfo(
        num_distinct_actions=self.network.num_actions(),
        max_chance_outcomes=0,
        num_players=len(self._vehicles),
        min_utility=-max_num_time_step - 1,
        max_utility=0,
        max_game_length=max_num_time_step)
    super().__init__(_GAME_TYPE, game_info, params if params else {})

  def new_initial_state(self) -> "DynamicRoutingGameState":
    """Returns the state corresponding to the start of a game."""
    return DynamicRoutingGameState(self, self._vehicles, self.time_step_length)

  def make_py_observer(self, iig_obs_type=None, params=None):
    """Returns a NetworkObserver object used for observing game state."""
    if ((iig_obs_type is None) or
        (iig_obs_type.public_info and not iig_obs_type.perfect_recall)):
      return NetworkObserver(self.num_players(), self.max_game_length())
    return IIGObserverForPublicInfoGame(iig_obs_type, params)


class DynamicRoutingGameState(pyspiel.State):
  """State of the DynamicRoutingGame.

  One player is equal to one vehicle.
  See docstring of the game class and of the file for more information.
  Attributes:
    _current_time_step: current time step of the game.
    _is_terminal: boolean that encodes weither the game is over.
    _time_step_length: size of the time step, used to convert travel times into
      number of game time steps.
    _vehicle_at_destination: set of vehicles that have reached their
      destinations. When a vehicle has reached its destination but the game is
      not finished, it cannot do anything.
    _vehicle_destinations: the destination of each vehicle.
    _vehicle_final_arrival_times: the arrival times of each vehicle, the arrival
      is either 0 if the vehicle is still in the network or its arrival time if
      the vehicle has reached its destination.
    _vehicle_locations: current location of the vehicles as a network road
      section.
    _vehicle_without_legal_actions: list of vehicles without legal actions at
      next time step. This is required because if no vehicle has legal actions
      for a simultaneous node then an error if raised.
    _waiting_times: time that each vehicle should wait before being able to move
      to the next road section.
  """
  _current_time_step: int
  _is_terminal: bool
  _time_step_length: float
  _vehicle_at_destination: Set[int]
  _vehicle_destinations: List[str]
  _vehicle_final_arrival_times: List[float]
  _vehicle_locations: List[str]
  _vehicle_without_legal_actions: Set[int]
  _waiting_times: List[int]

  def __init__(self, game: DynamicRoutingGame,
               vehicles: Iterable[dynamic_routing_utils.Vehicle],
               time_step_length: float):
    """Constructor; should only be called by Game.new_initial_state."""
    super().__init__(game)
    self._current_time_step = 0
    self._is_terminal = False
    self._time_step_length = time_step_length
    self._vehicle_at_destination = set()
    self._vehicle_destinations = [vehicle.destination for vehicle in vehicles]
    self._vehicle_final_arrival_times = [0.0 for _ in vehicles]
    self._vehicle_locations = [vehicle.origin for vehicle in vehicles]
    self._vehicle_without_legal_actions = set()
    self._waiting_times = [
        int(veh._departure_time / self._time_step_length) for veh in vehicles
    ]
    self.running_cost = [0 for vehicle in vehicles]

  @property
  def current_time_step(self) -> int:
    """Return current time step."""
    return self._current_time_step

  def current_player(self) -> pyspiel.PlayerId:
    """Returns the current player.

    If the game is over, TERMINAL is returned. If the game is at a chance
    node then CHANCE is returned. Otherwise SIMULTANEOUS is returned.
    """
    if self._is_terminal:
      return pyspiel.PlayerId.TERMINAL
    return pyspiel.PlayerId.SIMULTANEOUS

  def assert_valid_player(self, vehicle: int):
    """Assert that a vehicle as a int between 0 and num_players."""
    assert isinstance(vehicle, int), f"{vehicle} is not a int."
    assert vehicle >= 0, f"player: {vehicle}<0."
    assert vehicle < self.get_game().num_players(), (
        f"player: {vehicle} >= num_players: {self.get_game().num_players()}")

  def _legal_actions(self, vehicle: int) -> List[int]:
    """Return the legal actions of the vehicle.

    Legal actions are the succesor road section of the vehicle current road
    section.
    Args:
      vehicle: the vehicle id.

    Returns:
      list_legal_actions: a list of legal actions. If the game is finished then
        the list is empty. If the vehicle is at its destination, has a positive
        waiting time or if it is on a node without successors then an empty list
        is returned. Otherwise the list of successors nodes of the current
        vehicle location is returned.
    """
    if self._is_terminal:
      return []
    if self.get_game().perform_sanity_checks:
      self.assert_valid_player(vehicle)
    if vehicle in self._vehicle_without_legal_actions:
      # If the vehicle is at destination it cannot do anything.
      return [dynamic_routing_utils.NO_POSSIBLE_ACTION]
    if self._waiting_times[vehicle] > 0:
      return [dynamic_routing_utils.NO_POSSIBLE_ACTION]
    _, end_section_node = dynamic_routing_utils._nodes_from_road_section(  # pylint:disable=protected-access
        self._vehicle_locations[vehicle])
    successors = self.get_game().network.get_successors(end_section_node)
    if successors:
      assert isinstance(successors, Iterable)
      actions = [
          self.get_game().network.get_action_id_from_movement(
              end_section_node, d) for d in successors
      ]
      if self.get_game().perform_sanity_checks:
        map(self.get_game().network.assert_valid_action, actions)
      return sorted(actions)
    return []

  def _apply_actions(self, actions: List[int]):
    """Applies the specified action to the state.

    For each vehicle's action, if the vehicle is not at a sink node, if the
    action is valid and if the waiting time is negative, then the vehicle will
    move to the successor link corresponding to its action.
    The function then detects if the vehicle has reached its destination or
    a sink node and updates _vehicle_at_destination,
    _vehicle_without_legal_actions and _vehicle_final_arrival_times
    accordingly.
    The function then assigns waiting for each vehicle that have moved based on
    the new volume of cars on the link they reach.
    The function evolves the time and checks if the game is finished.
    Args:
        actions: the action chosen by each vehicle.
    """
    if self.get_game().perform_sanity_checks:
      assert not self._is_terminal
    if self.get_game().perform_sanity_checks:
      assert isinstance(actions, Iterable)
      assert len(actions) == self.get_game().num_players(), (
          f"Each player does not have an actions. Actions has {len(actions)} "
          f"elements, it should have {self.get_game().num_players()}.")
    for vehicle_id, action in enumerate(actions):
      if vehicle_id not in self._vehicle_at_destination:
        self.running_cost[vehicle_id] += self._time_step_length
      # Has the vehicle already reached a sink node?
      if vehicle_id in self._vehicle_without_legal_actions:
        if self.get_game().perform_sanity_checks:
          assert action == dynamic_routing_utils.NO_POSSIBLE_ACTION, (
              f"{action} should be {dynamic_routing_utils.NO_POSSIBLE_ACTION}.")
        continue
      if self._waiting_times[vehicle_id] > 0:
        continue
      if self.get_game().perform_sanity_checks:
        self.get_game().network.assert_valid_action(
            action, self._vehicle_locations[vehicle_id])
      self._vehicle_locations[vehicle_id] = (
          self.get_game().network.get_road_section_from_action_id(action))
      if (self._vehicle_locations[vehicle_id] ==
          self._vehicle_destinations[vehicle_id]):
        self._vehicle_final_arrival_times[vehicle_id] = self._current_time_step
        self._vehicle_at_destination.add(vehicle_id)
        self._vehicle_without_legal_actions.add(vehicle_id)
      # Will the vehicle have a legal action for next time step?
      elif self.get_game().network.is_location_at_sink_node(
          self._vehicle_locations[vehicle_id]):
        self._vehicle_without_legal_actions.add(vehicle_id)
    self._current_time_step += 1
    volumes = {}
    for road_section in self._vehicle_locations:
      if road_section not in volumes:
        volumes[road_section] = 0
      # Each vehicle has a weight a one.
      volumes[road_section] += 1
    for vehicle_id, _ in enumerate(actions):
      # Has the vehicle already reached a sink node?
      if vehicle_id in self._vehicle_without_legal_actions:
        continue
      if self._waiting_times[vehicle_id] > 0:
        self._waiting_times[vehicle_id] -= 1
      else:
        self._waiting_times[vehicle_id] = int(self.get_game(
        ).network.get_travel_time(self._vehicle_locations[vehicle_id], volumes[
            self._vehicle_locations[vehicle_id]]) / self._time_step_length -
                                              1.0)
    # Is the game finished?
    if (self._current_time_step >= self.get_game().max_game_length() or len(
        self._vehicle_without_legal_actions) == self.get_game().num_players()):
      self._is_terminal = True
      for vehicle_id in range(self.get_game().num_players()):
        if vehicle_id not in self._vehicle_at_destination:
          self._vehicle_final_arrival_times[vehicle_id] = (
              self._current_time_step)

  def _action_to_string(self, player, action) -> str:
    """Action -> string."""
    if self.get_game().perform_sanity_checks:
      self.assert_valid_player(player)
    if action == dynamic_routing_utils.NO_POSSIBLE_ACTION:
      return f"Vehicle {player} reach a sink node or its destination."
    if self.get_game().perform_sanity_checks:
      self.get_game().network.assert_valid_action(action)
    return (f"Vehicle {player} would like to move to "
            f"{self.get_game().network.get_road_section_from_action_id(action)}"
            ".")

  def is_terminal(self) -> bool:
    """Returns True if the game is over."""
    return self._is_terminal

  def rewards(self):
    """Reward at the previous step."""
    if self._is_terminal or self._current_time_step == 0:
      return [0 for _ in self._vehicle_locations]
    reward = [-self._time_step_length for _ in self._vehicle_locations]
    for vehicle in self._vehicle_at_destination:
      reward[vehicle] = 0
    return reward

  def returns(self) -> List[float]:
    """Total reward for each player over the course of the game so far."""
    if not self._is_terminal:
      returns = [
          -self._time_step_length * self.current_time_step
          for _ in self._vehicle_locations
      ]
      for vehicle in self._vehicle_at_destination:
        returns[vehicle] = -(
            self._vehicle_final_arrival_times[vehicle] * self._time_step_length)
      return returns
    returns = [
        -arrival_time * self._time_step_length
        for arrival_time in self._vehicle_final_arrival_times
    ]
    return returns

  def get_current_vehicle_locations(self) -> List[str]:
    """Get vehicle locations for debug purposes."""
    return self._vehicle_locations

  def get_location_as_int(self, vehicle: int) -> int:
    """Get the vehicle location."""
    origin, destination = dynamic_routing_utils._nodes_from_road_section(  # pylint:disable=protected-access
        self._vehicle_locations[vehicle])
    return self.get_game().network.get_action_id_from_movement(
        origin, destination)

  def get_current_vehicle_locations_as_int(self) -> List[int]:
    """Get locations of all vehicles for the observation tensor."""
    return [
        self.get_location_as_int(x)
        for x in range(self.get_game().num_players())
    ]

  def __str__(self) -> str:
    """String for debug purposes. No particular semantics are required."""
    if self._is_terminal:
      time = f"{self._current_time_step}, game finished."
    else:
      time = f"{self._current_time_step}"
    return (f"Vehicle locations: {self._vehicle_locations}, "
            f"time: {time}, waiting_time={self._waiting_times}.")


class NetworkObserver:
  """Network observer used by the learning algorithm.

  The state string is the state history string. The state tensor is an array
  of size max_game_length, num_players where each element is the location of
  the vehicle at this time.
  Attributes:
    dict: dictionary {"observation": tensor}.
    tensor: list of location for each time step.
  """

  def __init__(self, num_vehicles: int, num_time: int):
    """Initializes an empty observation tensor."""
    shape = (num_time + 1, num_vehicles + 1)
    self.tensor = np.zeros(np.prod(shape), np.float32)
    self.dict = {"observation": np.reshape(self.tensor, shape)}

  def set_from(self, state, player):
    """Update the state tensor.

    Put the locations of each players in the tensor row corresponding to
    the current time step. Insert the current player location at the
    beginning of the row.
    Args:
      state: the state,
      player: the player.
    """
    vehicles = state.get_current_vehicle_locations_as_int()
    vehicles.insert(0, state.get_location_as_int(player))
    self.dict["observation"][state.current_time_step, :] = vehicles

  def string_from(self, state, player):
    """Return the state history string."""
    return f"{player}: {state.history_str()}"


# Register the game with the OpenSpiel library
pyspiel.register_game(_GAME_TYPE, DynamicRoutingGame)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Example use of the NeuRD algorithm on Kuhn Poker.

This NeuRD implementation does not use an entropy bonus to ensure that the
current joint policy approaches an equilibrium in zero-sum games, but it
tracks the exact tabular average so that the average policy approaches an
equilibrium (assuming the policy networks train well).
"""

from absl import app
from absl import flags
import tensorflow.compat.v1 as tf

from open_spiel.python.algorithms import neurd
import pyspiel

tf.enable_eager_execution()

FLAGS = flags.FLAGS

flags.DEFINE_integer("iterations", 1000, "Number of iterations")
flags.DEFINE_string("game", "kuhn_poker", "Name of the game")
flags.DEFINE_integer("players", 2, "Number of players")
flags.DEFINE_integer("print_freq", 100, "How often to print the exploitability")
flags.DEFINE_integer("num_hidden_layers", 1,
                     "The number of hidden layers in the policy model.")
flags.DEFINE_integer("num_hidden_units", 13,
                     "The number of hidden layers in the policy model.")
flags.DEFINE_integer(
    "num_hidden_factors", 8,
    "The number of factors in each hidden layer in the policy model.")
flags.DEFINE_boolean(
    "use_skip_connections", True,
    "Whether or not to use skip connections in the policy model.")
flags.DEFINE_integer("batch_size", 100, "The policy model training batch size.")
flags.DEFINE_float(
    "threshold", 2.,
    "Logits of the policy model will be discouraged from growing beyond "
    "`threshold`.")
flags.DEFINE_float("step_size", 1.0, "Policy model step size.")
flags.DEFINE_boolean(
    "autoencode", False,
    "Whether or not to augment the policy model with outputs that attempt to "
    "reproduce the model inputs. The policy model is updated online so "
    "training with the reproduction error as an auxiliary task helps to keep "
    "the model stable in the absence of an entropy bonus.")


def main(_):
  game = pyspiel.load_game(FLAGS.game, {"players": FLAGS.players})

  models = []
  for _ in range(game.num_players()):
    models.append(
        neurd.DeepNeurdModel(
            game,
            num_hidden_layers=FLAGS.num_hidden_layers,
            num_hidden_units=FLAGS.num_hidden_units,
            num_hidden_factors=FLAGS.num_hidden_factors,
            use_skip_connections=FLAGS.use_skip_connections,
            autoencode=FLAGS.autoencode))

  solver = neurd.CounterfactualNeurdSolver(game, models)

  def _train(model, data):
    neurd.train(
        model,
        data,
        batch_size=FLAGS.batch_size,
        step_size=FLAGS.step_size,
        threshold=FLAGS.threshold,
        autoencoder_loss=(tf.compat.v1.losses.huber_loss
                          if FLAGS.autoencode else None))

  for i in range(FLAGS.iterations):
    solver.evaluate_and_update_policy(_train)
    if i % FLAGS.print_freq == 0:
      conv = pyspiel.exploitability(game, solver.average_policy())
      print("Iteration {} exploitability {}".format(i, conv))


if __name__ == "__main__":
  app.run(main)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Simple network classes for Tensorflow based on tf.Module."""

import math
import tensorflow.compat.v1 as tf

# Temporarily disable TF2 behavior until code is updated.
tf.disable_v2_behavior()

# This code is based directly on the TF docs:
# https://www.tensorflow.org/versions/r1.15/api_docs/python/tf/Module


class Linear(tf.Module):
  """A simple linear module.

  Always includes biases and only supports ReLU activations.
  """

  def __init__(self, in_size, out_size, activate_relu=True, name=None):
    """Creates a linear layer.

    Args:
      in_size: (int) number of inputs
      out_size: (int) number of outputs
      activate_relu: (bool) whether to include a ReLU activation layer
      name: (string): the name to give to this layer
    """

    super(Linear, self).__init__(name=name)
    self._activate_relu = activate_relu
    # Weight initialization inspired by Sonnet's Linear layer,
    # which cites https://arxiv.org/abs/1502.03167v3
    stddev = 1.0 / math.sqrt(in_size)
    self._weights = tf.Variable(
        tf.random.truncated_normal([in_size, out_size], mean=0.0,
                                   stddev=stddev),
        name="weights")
    self._bias = tf.Variable(tf.zeros([out_size]), name="bias")

  def __call__(self, tensor):
    y = tf.matmul(tensor, self._weights) + self._bias
    return tf.nn.relu(y) if self._activate_relu else y


class Sequential(tf.Module):
  """A simple sequential module.

  Always includes biases and only supports ReLU activations.
  """

  def __init__(self, layers, name=None):
    """Creates a model from successively applying layers.

    Args:
      layers: Iterable[tf.Module] that can be applied.
      name: (string): the name to give to this layer
    """

    super(Sequential, self).__init__(name=name)
    self._layers = layers

  def __call__(self, tensor):
    for layer in self._layers:
      tensor = layer(tensor)
    return tensor


class MLP(tf.Module):
  """A simple dense network built from linear layers above."""

  def __init__(self,
               input_size,
               hidden_sizes,
               output_size,
               activate_final=False,
               name=None):
    """Create the MLP.

    Args:
      input_size: (int) number of inputs
      hidden_sizes: (list) sizes (number of units) of each hidden layer
      output_size: (int) number of outputs
      activate_final: (bool) should final layer should include a ReLU
      name: (string): the name to give to this network
    """

    super(MLP, self).__init__(name=name)
    self._layers = []
    with self.name_scope:
      # Hidden layers
      for size in hidden_sizes:
        self._layers.append(Linear(in_size=input_size, out_size=size))
        input_size = size
      # Output layer
      self._layers.append(
          Linear(
              in_size=input_size,
              out_size=output_size,
              activate_relu=activate_final))

  @tf.Module.with_name_scope
  def __call__(self, x):
    for layer in self._layers:
      x = layer(x)
    return x


class MLPTorso(tf.Module):
  """A specialized half-MLP module when constructing multiple heads.

  Note that every layer includes a ReLU non-linearity activation.
  """

  def __init__(self, input_size, hidden_sizes, name=None):
    super(MLPTorso, self).__init__(name=name)
    self._layers = []
    with self.name_scope:
      for size in hidden_sizes:
        self._layers.append(Linear(in_size=input_size, out_size=size))
        input_size = size

  @tf.Module.with_name_scope
  def __call__(self, x):
    for layer in self._layers:
      x = layer(x)
    return x



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Implementation of Nash Conv metric for a policy.

In the context of mean field games, the Nash Conv is the difference between:
- the value of a policy against the distribution of that policy,
- and the best response against the distribution of the policy.
"""

from open_spiel.python import policy as policy_std
from open_spiel.python.mfg import value
from open_spiel.python.mfg.algorithms import best_response_value
from open_spiel.python.mfg.algorithms import distribution
from open_spiel.python.mfg.algorithms import policy_value


class NashConv(object):
  """Computes the Nash Conv of a policy."""

  def __init__(self, game, policy: policy_std.Policy, root_state=None):
    """Initializes the nash conv.

    Args:
      game: The game to analyze.
      policy: A `policy.Policy` object.
      root_state: The state of the game at which to start. If `None`, the game
        root state is used.
    """
    self._game = game
    self._policy = policy
    if root_state is None:
      self._root_states = game.new_initial_states()
    else:
      self._root_states = [root_state]
    self._distrib = distribution.DistributionPolicy(
        self._game, self._policy, root_state=root_state)
    self._pi_value = policy_value.PolicyValue(
        self._game,
        self._distrib,
        self._policy,
        value.TabularValueFunction(self._game),
        root_state=root_state)
    self._br_value = best_response_value.BestResponse(
        self._game,
        self._distrib,
        value.TabularValueFunction(self._game),
        root_state=root_state)

  def nash_conv(self):
    """Returns the nash conv.

    Returns:
      A float representing the nash conv for the policy.
    """
    return sum([
        self._br_value.eval_state(state) - self._pi_value.eval_state(state)
        for state in self._root_states
    ])

  def br_values(self):
    """Returns the best response values to the policy distribution.

    Returns:
      A List[float] representing the best response values for a policy
        distribution.
    """
    return [self._br_value.eval_state(state) for state in self._root_states]

  @property
  def distribution(self) -> distribution.DistributionPolicy:
    return self._distrib



# Copyright 2022 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Mean Field Crowd Modelling Game in 2d.

Please see the C++ implementation under games/mfg/crowd_modelling_2d.h for
more information.
"""

from typing import Sequence


def grid_to_forbidden_states(grid: Sequence[str]) -> str:
  """Converts a grid into string representation of forbidden states.

  Args:
    grid: Rows of the grid. '#' character denotes a forbidden state. All rows
      should have the same number of columns, i.e. cells.

  Returns:
    String representation of forbidden states in the form of x (column) and y
    (row) pairs, e.g. [1|1;0|2].
  """
  forbidden_states = []
  num_cols = len(grid[0])
  for y, row in enumerate(grid):
    assert len(row) == num_cols, f'Number of columns should be {num_cols}.'
    for x, cell in enumerate(row):
      if cell == '#':
        forbidden_states.append(f'{x}|{y}')
  return '[' + ';'.join(forbidden_states) + ']'


FOUR_ROOMS_FORBIDDEN_STATES = grid_to_forbidden_states([
    '#############',
    '#     #     #',
    '#     #     #',
    '#           #',
    '#     #     #',
    '#     #     #',
    '### ##### ###',
    '#     #     #',
    '#     #     #',
    '#           #',
    '#     #     #',
    '#     #     #',
    '#############',
])

# Four rooms with an initial state at top-left corner.
FOUR_ROOMS = {
    'forbidden_states': FOUR_ROOMS_FORBIDDEN_STATES,
    'horizon': 40,
    'initial_distribution': '[1|1]',
    'initial_distribution_value': '[1.0]',
    'size': 13,
}

MAZE_FORBIDDEN_STATES = grid_to_forbidden_states([
    '######################',
    '#      #     #     # #',
    '#      #     #     # #',
    '######    #  # ##  # #',
    '#         #  # #   # #',
    '#         #  # ### # #',
    '#  ########  #   #   #',
    '#    # # #  ##   #   #',
    '#    # # #     # # ###',
    '#    # # #     # # # #',
    '###### # ####### # # #',
    '#  #         #   # # #',
    '#  # ## ###  #   # # #',
    '## # #    #  ##### # #',
    '## # # #  #      # # #',
    '#    # ####        # #',
    '# ####  # ########   #',
    '#       #  #   # ### #',
    '#  #  # #  # # #   # #',
    '# ##### #    # #     #',
    '#            #       #',
    '######################',
])

# 22x22 maze with an initial state at top-left corner,
MAZE = {
    'forbidden_states': MAZE_FORBIDDEN_STATES,
    'horizon': 100,
    'initial_distribution': '[1|1]',
    'initial_distribution_value': '[1.0]',
    'size': 22,
}



# Copyright 2022 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests counterfactual regret minimization."""

from absl.testing import absltest
from absl.testing import parameterized

from open_spiel.python.examples.meta_cfr.sequential_games import cfr
from open_spiel.python.examples.meta_cfr.sequential_games import game_tree_utils as trees
from open_spiel.python.examples.meta_cfr.sequential_games import openspiel_api


def _uniform_policy(size):
  if size > 0:
    return [1./size]*size
  return []


class CfrTest(parameterized.TestCase):

  @parameterized.named_parameters(('kuhn_poker_test', 'kuhn_poker'),
                                  ('leduc_poker_test', 'leduc_poker'))
  def test_zero_policy_is_uniform(self, game):
    config = {'players': 2}
    cfr_game_tree = trees.build_game_tree(
        openspiel_api.WorldState(
            game_name=game, config=config, perturbation=False))
    cfr.compute_cfr_values(cfr_game_tree, 1)
    infostates_p1 = list(cfr_game_tree.all_infostates_map[1].values())
    infostates_p2 = list(cfr_game_tree.all_infostates_map[2].values())
    with self.subTest('player_1_initial_policy'):
      for i in range(len(infostates_p1)):
        self.assertListEqual(
            list(infostates_p1[i].policy.values()),
            _uniform_policy(len(infostates_p1[i].policy.values())))
    with self.subTest('player_2_initial_policy'):
      for i in range(len(infostates_p2)):
        self.assertListEqual(
            list(infostates_p2[i].policy.values()),
            _uniform_policy(len(infostates_p2[i].policy.values())))

  def test_cfr_leduc_poker(self):
    config = {'players': 2}
    exploitability_error = 0.2
    cfr_game_tree = trees.build_game_tree(
        openspiel_api.WorldState(
            game_name='leduc_poker', config=config, perturbation=False))
    best_response_value_p1, best_response_value_p2 = cfr.compute_cfr_values(
        cfr_game_tree, 20)
    last_best_response_value_player_1 = best_response_value_p1[-1]
    last_best_response_value_player_2 = best_response_value_p2[-1]
    exploitability = (last_best_response_value_player_1 +
                      last_best_response_value_player_2) / 2
    # Exploitability values are computed using OpenSpiel cfr
    self.assertLessEqual(exploitability, 0.59 + exploitability_error)

  def test_cfr_kuhn_poker(self):
    config = {'players': 2}
    exploitability_error = 0.2
    cfr_game_tree = trees.build_game_tree(
        openspiel_api.WorldState(
            game_name='kuhn_poker', config=config, perturbation=False))
    best_response_value_p1, best_response_value_p2 = cfr.compute_cfr_values(
        cfr_game_tree, 20)
    last_best_response_value_player_1 = best_response_value_p1[-1]
    last_best_response_value_player_2 = best_response_value_p2[-1]
    exploitability = (last_best_response_value_player_1 +
                      last_best_response_value_player_2) / 2
    # Exploitability values are computed using OpenSpiel cfr
    self.assertLessEqual(exploitability, 0.06 + exploitability_error)


if __name__ == '__main__':
  absltest.main()



# Copyright 2022 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Simple example of using the Roshambo population.

Note: the Roshambo bots are an optional dependency and excluded by default.
To enable Roshambo bots, set OPEN_SPIEL_BUILD_WITH_ROSHAMBO to ON when building.
See
https://github.com/deepmind/open_spiel/blob/master/docs/install.md#configuring-conditional-dependencies
for details.
"""

import re
from absl import app
from absl import flags
import numpy as np

from open_spiel.python import games  # pylint: disable=unused-import
from open_spiel.python import rl_agent
from open_spiel.python import rl_environment
import pyspiel

FLAGS = flags.FLAGS

# See open_spiel/data/paper_data/pbe_rrps for the bot table from the RRPS paper:
# https://arxiv.org/abs/2303.03196
flags.DEFINE_string("bot_table_file", None,
                    "The file containing the bot entries.")

flags.DEFINE_integer("player0_pop_id", 0, "Population member ID for player 0")
flags.DEFINE_integer("player1_pop_id", 1, "Population member ID for player 1")
flags.DEFINE_integer("seed", 0, "Seed to use for RNG")
flags.DEFINE_integer("env_recall", 1,
                     "Number of recent steps to include in observation")


class BotAgent(rl_agent.AbstractAgent):
  """Agent class that wraps a bot.

  Note, the environment must include the OpenSpiel state in its observations,
  which means it must have been created with use_full_state=True.
  """

  def __init__(self, num_actions, bot, name="bot_agent"):
    assert num_actions > 0
    self._bot = bot
    self._num_actions = num_actions

  def restart(self):
    self._bot.restart()

  def step(self, time_step, is_evaluation=False):
    # If it is the end of the episode, don't select an action.
    if time_step.last():
      return

    _, state = pyspiel.deserialize_game_and_state(
        time_step.observations["serialized_state"])

    action = self._bot.step(state)
    probs = np.zeros(self._num_actions)
    probs[action] = 1.0

    return rl_agent.StepOutput(action=action, probs=probs)


def eval_agents(env, agents, num_players, num_episodes):
  """Evaluate the agent."""
  sum_episode_rewards = np.zeros(num_players)
  for ep in range(num_episodes):
    for agent in agents:
      # Bots need to be restarted at the start of the episode.
      if hasattr(agent, "restart"):
        agent.restart()
    time_step = env.reset()
    episode_rewards = np.zeros(num_players)
    while not time_step.last():
      agents_output = [
          agent.step(time_step, is_evaluation=True) for agent in agents
      ]
      action_list = [agent_output.action for agent_output in agents_output]
      time_step = env.step(action_list)
      episode_rewards += time_step.rewards
    sum_episode_rewards += episode_rewards
    print(f"Finished episode {ep}, "
          + f"avg returns: {sum_episode_rewards / num_episodes}")

  return sum_episode_rewards / num_episodes


def print_roshambo_bot_names_and_ids(roshambo_bot_names):
  print("Roshambo bot population:")
  for i in range(len(roshambo_bot_names)):
    print(f"{i}: {roshambo_bot_names[i]}")


def create_roshambo_bot_agent(player_id, num_actions, bot_names, pop_id):
  name = bot_names[pop_id]
  # Creates an OpenSpiel bot with the default number of throws
  # (pyspiel.ROSHAMBO_NUM_THROWS). To create one for a different number of
  # throws per episode, add the number as the third argument here.
  bot = pyspiel.make_roshambo_bot(player_id, name)
  return BotAgent(num_actions, bot, name=name)


def analyze_bot_table(filename):
  """Do some analysis on the payoff cross-table."""
  print(f"Opening bot table file: {filename}")
  bot_table_file = open(filename, "r")
  table = np.zeros(shape=(pyspiel.ROSHAMBO_NUM_BOTS,
                          pyspiel.ROSHAMBO_NUM_BOTS), dtype=np.float64)
  print("Parsing file...")
  values = {}
  bot_names_map = {}
  for line in bot_table_file:
    line = line.strip()
    # ('driftbot', 'driftbot', -0.571)
    myre = re.compile(r"\'(.*)\', \'(.*)\', (.*)\)")
    match_obj = myre.search(line)
    row_agent, col_agent, value = match_obj.groups()
    values[f"{row_agent},{col_agent}"] = value
    bot_names_map[row_agent] = True
  bot_names_list = list(bot_names_map.keys())
  bot_names_list.sort()
  print(len(bot_names_list))
  assert len(bot_names_list) == pyspiel.ROSHAMBO_NUM_BOTS
  print(bot_names_list)
  for i in range(pyspiel.ROSHAMBO_NUM_BOTS):
    for j in range(pyspiel.ROSHAMBO_NUM_BOTS):
      key = f"{bot_names_list[i]},{bot_names_list[j]}"
      assert key in values
      table[i][j] = float(values[key])
  print("Population returns:")
  pop_returns = np.zeros(pyspiel.ROSHAMBO_NUM_BOTS)
  pop_aggregate = np.zeros(pyspiel.ROSHAMBO_NUM_BOTS)
  for i in range(pyspiel.ROSHAMBO_NUM_BOTS):
    pop_eval = 0
    for j in range(pyspiel.ROSHAMBO_NUM_BOTS):
      pop_eval += table[i][j]
    pop_eval /= pyspiel.ROSHAMBO_NUM_BOTS
    # print(f"  {bot_names_list[i]}: {pop_eval}")
    pop_returns[i] = pop_eval
    pop_aggregate[i] += pop_eval
    print(f"  {pop_eval},")
  print("Population exploitabilities: ")
  pop_expls = np.zeros(pyspiel.ROSHAMBO_NUM_BOTS)
  avg_pop_expl = 0
  for i in range(pyspiel.ROSHAMBO_NUM_BOTS):
    pop_expl = -float(pyspiel.ROSHAMBO_NUM_THROWS)
    for j in range(pyspiel.ROSHAMBO_NUM_BOTS):
      pop_expl = max(pop_expl, -table[i][j])
    avg_pop_expl += pop_expl
    pop_expls[i] = pop_expl
    pop_aggregate[i] -= pop_expl
    print(f"  {pop_expl},")
  avg_pop_expl /= pyspiel.ROSHAMBO_NUM_BOTS
  print(f"Avg within-pop expl: {avg_pop_expl}")
  print("Aggregate: ")
  indices = np.argsort(pop_aggregate)
  for i in range(pyspiel.ROSHAMBO_NUM_BOTS):
    idx = indices[pyspiel.ROSHAMBO_NUM_BOTS - i - 1]
    print(f"  {i+1} & \\textsc{{{bot_names_list[idx]}}} & " +
          f" ${pop_returns[idx]:0.3f}$ " +
          f"& ${pop_expls[idx]:0.3f}$ & ${pop_aggregate[idx]:0.3f}$ \\\\")
  print("Dominance:")
  for i in range(pyspiel.ROSHAMBO_NUM_BOTS):
    for j in range(pyspiel.ROSHAMBO_NUM_BOTS):
      if np.all(np.greater(table[i], table[j])):
        print(f"{bot_names_list[i]} dominates {bot_names_list[j]}")


def main(_):
  np.random.seed(FLAGS.seed)

  if FLAGS.bot_table_file is not None:
    analyze_bot_table(FLAGS.bot_table_file)
    return

  # Note that the include_full_state variable has to be enabled because the
  # BotAgent needs access to the full state.
  env = rl_environment.Environment(
      "repeated_game(stage_game=matrix_rps(),num_repetitions=" +
      f"{pyspiel.ROSHAMBO_NUM_THROWS}," +
      f"recall={FLAGS.env_recall})",
      include_full_state=True)
  num_players = 2
  num_actions = env.action_spec()["num_actions"]
  # Learning agents might need this:
  # info_state_size = env.observation_spec()["info_state"][0]

  print("Loading population...")
  pop_size = pyspiel.ROSHAMBO_NUM_BOTS
  print(f"Population size: {pop_size}")
  roshambo_bot_names = pyspiel.roshambo_bot_names()
  roshambo_bot_names.sort()
  print_roshambo_bot_names_and_ids(roshambo_bot_names)

  bot_id = 0
  roshambo_bot_ids = {}
  for name in roshambo_bot_names:
    roshambo_bot_ids[name] = bot_id
    bot_id += 1

  # Create two bot agents
  agents = [
      create_roshambo_bot_agent(0, num_actions, roshambo_bot_names,
                                FLAGS.player0_pop_id),
      create_roshambo_bot_agent(1, num_actions, roshambo_bot_names,
                                FLAGS.player1_pop_id)
  ]

  print("Starting eval run.")
  print(f"Player 0 is (pop_id {FLAGS.player0_pop_id}: " +
        f"{roshambo_bot_names[FLAGS.player0_pop_id]})")
  print(f"Player 1 is (pop_id {FLAGS.player1_pop_id}: " +
        f"{roshambo_bot_names[FLAGS.player1_pop_id]})")
  avg_eval_returns = eval_agents(env, agents, num_players, 100)
  print(avg_eval_returns)


if __name__ == "__main__":
  app.run(main)



# Copyright 2022 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A vectorized RL Environment."""


class SyncVectorEnv(object):
  """A vectorized RL Environment.

  This environment is synchronized - games do not execute in parallel. Speedups
  are realized by calling models on many game states simultaneously.
  """

  def __init__(self, envs):
    if not isinstance(envs, list):
      raise ValueError(
          "Need to call this with a list of rl_environment.Environment objects")
    self.envs = envs

  def __len__(self):
    return len(self.envs)

  def observation_spec(self):
    return self.envs[0].observation_spec()

  @property
  def num_players(self):
    return self.envs[0].num_players

  def step(self, step_outputs, reset_if_done=False):
    """Apply one step.

    Args:
      step_outputs: the step outputs
      reset_if_done: if True, automatically reset the environment
          when the epsiode ends

    Returns:
      time_steps: the time steps,
      reward: the reward
      done: done flag
      unreset_time_steps: unreset time steps
    """
    time_steps = [
        self.envs[i].step([step_outputs[i].action])
        for i in range(len(self.envs))
    ]
    reward = [step.rewards for step in time_steps]
    done = [step.last() for step in time_steps]
    unreset_time_steps = time_steps  # Copy these because you may want to look
                                     # at the unreset versions to extract
                                     # information from them

    if reset_if_done:
      time_steps = self.reset(envs_to_reset=done)

    return time_steps, reward, done, unreset_time_steps

  def reset(self, envs_to_reset=None):
    if envs_to_reset is None:
      envs_to_reset = [True for _ in range(len(self.envs))]

    time_steps = [
        self.envs[i].reset()
        if envs_to_reset[i] else self.envs[i].get_time_step()
        for i in range(len(self.envs))
    ]
    return time_steps



# Copyright 2022 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for open_spiel.python.sequence_form_utils.py."""

from absl.testing import absltest
from absl.testing import parameterized

import numpy as np

from open_spiel.python import policy
from open_spiel.python.algorithms import cfr
from open_spiel.python.algorithms import sequence_form_utils
from open_spiel.python.algorithms.expected_game_score import policy_value
import pyspiel

_KUHN_GAME = pyspiel.load_game('kuhn_poker')
_LEDUC_GAME = pyspiel.load_game('leduc_poker')


class SequenceFormTest(parameterized.TestCase):

  @parameterized.parameters(
      {
          'game': _KUHN_GAME,
          'cfr_iters': 100
      },
      {
          'game': _LEDUC_GAME,
          'cfr_iters': 10
      },
  )
  def test_sequence_to_policy(self, game, cfr_iters):

    cfr_solver = cfr.CFRSolver(game)

    for _ in range(cfr_iters):
      cfr_solver.evaluate_and_update_policy()

    (_, infoset_actions_to_seq, infoset_action_maps, _, _,
     _) = sequence_form_utils.construct_vars(game)

    policies = cfr_solver.average_policy()
    sequences = sequence_form_utils.policy_to_sequence(game, policies,
                                                       infoset_actions_to_seq)
    converted_policies = sequence_form_utils.sequence_to_policy(
        sequences, game, infoset_actions_to_seq, infoset_action_maps)
    np.testing.assert_allclose(
        policies.action_probability_array,
        converted_policies.action_probability_array,
        rtol=1e-10)

  @parameterized.parameters(
      {
          'game': _KUHN_GAME,
          'cfr_iters': 100
      },
      {
          'game': _LEDUC_GAME,
          'cfr_iters': 10
      },
  )
  def test_sequence_payoff(self, game, cfr_iters):
    (_, infoset_actions_to_seq, _, _, payoff_mat,
     _) = sequence_form_utils.construct_vars(game)

    uniform_policies = policy.TabularPolicy(game)
    uniform_value = policy_value(game.new_initial_state(),
                                 [uniform_policies, uniform_policies])
    sequences = sequence_form_utils.policy_to_sequence(game, uniform_policies,
                                                       infoset_actions_to_seq)
    np.testing.assert_allclose(
        uniform_value[0],
        -sequences[0].T @ payoff_mat @ sequences[1],
        rtol=1e-10)

    # use cfr iterations to construct new policy
    cfr_solver = cfr.CFRSolver(game)
    for _ in range(cfr_iters):
      cfr_solver.evaluate_and_update_policy()

    policies = cfr_solver.average_policy()
    cfr_value = policy_value(game.new_initial_state(), [policies, policies])
    sequences = sequence_form_utils.policy_to_sequence(game, policies,
                                                       infoset_actions_to_seq)
    np.testing.assert_allclose(
        cfr_value[0], -sequences[0].T @ payoff_mat @ sequences[1], rtol=1e-10)


if __name__ == '__main__':
  absltest.main()



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for open_spiel.python.algorithms.mip_nash."""

from absl.testing import absltest
import numpy as np

from open_spiel.python.algorithms.mip_nash import mip_nash
import pyspiel


class MIPNash(absltest.TestCase):

  def test_simple_games(self):
    # prisoners' dilemma
    pd_game = pyspiel.create_matrix_game(
        [[-2.0, -10.0], [0.0, -5.0]], [[-2.0, 0.0], [-10.0, -5.0]]
    )

    pd_eq = (np.array([0, 1]), np.array([0, 1]))

    computed_eq = mip_nash(pd_game, objective="MAX_SOCIAL_WELFARE")
    with self.subTest("pd"):
      np.testing.assert_array_almost_equal(computed_eq[0], pd_eq[0])
      np.testing.assert_array_almost_equal(computed_eq[1], pd_eq[1])

    # stag hunt
    sh_game = pyspiel.create_matrix_game(
        [[10.0, 1.0], [8.0, 5.0]], [[10.0, 8.0], [1.0, 5.0]]
    )

    sh_eq = (np.array([1, 0]), np.array([1, 0]))

    computed_eq = mip_nash(sh_game, objective="MAX_SOCIAL_WELFARE")
    with self.subTest("sh"):
      np.testing.assert_array_almost_equal(computed_eq[0], sh_eq[0])
      np.testing.assert_array_almost_equal(computed_eq[1], sh_eq[1])


if __name__ == "__main__":
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""An observation of a game.

This is intended to be the main way to get observations of states in Python.
The usage pattern is as follows:

0. Create the game we will be playing
1. Create each kind of observation required, using `make_observation`
2. Every time a new observation is required, call:
      `observation.set_from(state, player)`
   The tensor contained in the Observation class will be updated with an
   observation of the supplied state. This tensor is updated in-place, so if
   you wish to retain it, you must make a copy.

The following options are available when creating an Observation:
 - perfect_recall: if true, each observation must allow the observing player to
   reconstruct their history of actions and observations.
 - public_info: if true, the observation should include public information
 - private_info: specifies for which players private information should be
   included - all players, the observing player, or no players
 - params: game-specific parameters for observations

We ultimately aim to have all games support all combinations of these arguments.
However, initially many games will only support the combinations corresponding
to ObservationTensor and InformationStateTensor:
 - ObservationTensor: perfect_recall=False, public_info=True,
   private_info=SinglePlayer
 - InformationStateTensor: perfect_recall=True, public_info=True,
   private_info=SinglePlayer

Three formats of observation are supported:
a. 1-D numpy array, accessed by `observation.tensor`
b. Dict of numpy arrays, accessed by `observation.dict`. These are pieces of the
   1-D array, reshaped. The np.array objects refer to the same memory as the
   1-D array (no copying!).
c. String, hopefully human-readable (primarily for debugging purposes)

For usage examples, see `observation_test.py`.
"""

import numpy as np

import pyspiel


# Corresponds to the old information_state_XXX methods.
INFO_STATE_OBS_TYPE = pyspiel.IIGObservationType(perfect_recall=True)


class _Observation:
  """Contains an observation from a game."""

  def __init__(self, game, observer):
    self._observation = pyspiel._Observation(game, observer)
    self.dict = {}
    if self._observation.has_tensor():
      self.tensor = np.frombuffer(self._observation, np.float32)
      offset = 0
      for tensor_info in self._observation.tensors_info():
        size = np.prod(tensor_info.shape, dtype=np.int64)
        values = self.tensor[offset:offset + size].reshape(tensor_info.shape)
        self.dict[tensor_info.name] = values
        offset += size
    else:
      self.tensor = None

  def set_from(self, state, player):
    self._observation.set_from(state, player)

  def string_from(self, state, player):
    return (self._observation.string_from(state, player)
            if self._observation.has_string() else None)

  def compress(self):
    return self._observation.compress()

  def decompress(self, compressed_observation):
    self._observation.decompress(compressed_observation)


def make_observation(
    game,
    imperfect_information_observation_type=None,
    params=None,
):
  """Returns an _Observation instance if the imperfect_information_observation_type is supported, otherwise None."""
  params = params or {}
  if hasattr(game, 'make_py_observer'):
    return game.make_py_observer(imperfect_information_observation_type, params)
  else:
    if imperfect_information_observation_type is not None:
      observer = game.make_observer(
          imperfect_information_observation_type, params
      )
    else:
      observer = game.make_observer(params)
    if observer is None:
      return None
    return _Observation(game, observer)


class IIGObserverForPublicInfoGame:
  """Observer for imperfect information obvservations of public-info games."""

  def __init__(self, iig_obs_type, params):
    if params:
      raise ValueError(f'Observation parameters not supported; passed {params}')
    self._iig_obs_type = iig_obs_type
    self.tensor = None
    self.dict = {}

  def set_from(self, state, player):
    pass

  def string_from(self, state, player):
    del player
    if self._iig_obs_type.public_info:
      return state.history_str()
    else:
      return ''  # No private information to return
